# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Fixtures in order to test the SQLAlchemy backend."""

import copy
import datetime
import os

import fauxfactory
import pytest
from pytest_factoryboy import register
from unittest.mock import patch, PropertyMock

from sqlalchemy import create_engine, event

from nark.backends.sqlalchemy import objects
from nark.backends.sqlalchemy.storage import SQLAlchemyStore
from nark.items.activity import Activity
from nark.items.category import Category
from nark.items.fact import Fact
from nark.items.tag import Tag

from . import common, factories

register(factories.AlchemyCategoryFactory)
register(factories.AlchemyActivityFactory)
register(factories.AlchemyTagFactory)
register(factories.AlchemyFactFactory)


# *** SQLAlchemy fixtures


def implement_transactional_support_fully(engine):
    """
    (lb): A hack to make SQLite SAVEPOINTs work, so that
    session.begin_nested() works, so that if code under
    test calls session.commit(), we can still rollback().
    See the madness explained at:

      http://docs.sqlalchemy.org/en/rel_1_0/dialects/sqlite.html#pysqlite-serializable
    """

    @event.listens_for(engine, "connect")
    def do_connect(dbapi_connection, connection_record):
        # disable pysqlite's emitting of the BEGIN statement entirely.
        # also stops it from emitting COMMIT before any DDL.
        dbapi_connection.isolation_level = None

    @event.listens_for(engine, "begin")
    def do_begin(conn):
        # emit our own BEGIN
        conn.execute("BEGIN")


# (lb): Use 'module' scope so that the Session is only configured once
# for all tests. (Note: I also tried `scope='session'`, which seemed to
# work similarly.)
@pytest.fixture(scope='module')
def alchemy_runner(request):
    """
    Bind an in-memory mock-db to the session object.

    The session object, common.Session, is a global that this module-scoped
    function sets up just once. And then for each test, we'll create a clean
    session to use for testing that we clean up after.

    This is pretty much straight from the factoryboi docs:

      https://factoryboy.readthedocs.io/en/latest/orms.html#sqlalchemy

    """
    engine = create_engine('sqlite:///:memory:')
    implement_transactional_support_fully(engine)
    objects.metadata.bind = engine
    objects.metadata.create_all(engine)
    common.Session.configure(bind=engine)


@pytest.fixture
def alchemy_session(request):
    """
    Create a new session for each test.

    Reset the database and remove it after each test.

    As suggested by the docs:

        [Using factory_boy with ORMs](
            http://factoryboy.readthedocs.org/en/latest/orms.html#sqlalchemy
        )
    """
    # (lb): 2018-08-22: If we don't create a new Session() for each test,
    # there's one test that triggers a warning, TestFactManager.test_get_all:
    #   /home/user/.virtualenvs/dob/lib/python3.5/site-packages/
    #     sqlalchemy/orm/scoping.py:102: SAWarning: At least one scoped session
    #     is already present.  configure() can not affect sessions that have
    #     already been created.
    # Prepare a new, clean session for each test, lest SQLAlchemy complain.
    session = common.Session()

    # Rather than mock session.commit, use a nested transaction, otherwise
    # items' PK attrs will not be updated.
    # (lb): NOTE: I tried using subtransactions but those didn't work (albeit
    #   that was before I added the implement_transactional_support_fully hack).
    #       session.begin(subtransactions=True)  # (lb): Didn't work for me.
    # This works, but only after having hacked pysqlite to emit BEGIN ourselves.
    session.begin_nested()

    def fin():
        # "Rollback the session => no changes to the database."
        session.rollback()
        # (lb): Call rollback() twice. (For some reason one test,
        # TestFactManager.test_get_all, has five of each item in
        # the db after it runs, and after we can rollback once.
        # Which to me doesn't make sense. But whatever I'm only
        # human, I don't need to understand _everything_.)
        #
        # You can see for yourself with a conditional breakpoint, e.g.,
        #
        #   if session.query('name from categories;').all(): import pdb;pdb.set_trace()
        #
        # (lb): I'm not sure that we need to close, but so far hasn't hurt.
        session.close()
        # "Remove it, so that the next test gets a new Session()."
        common.Session.remove()

    request.addfinalizer(fin)

    return session


@pytest.fixture(params=[
    fauxfactory.gen_utf8(),
    fauxfactory.gen_alphanumeric(),
    ':memory:',
])
def db_path_parametrized(request, tmpdir):
    """Parametrized database paths."""
    if request.param == ':memory:':
        path = request.param
    else:
        path = os.path.join(tmpdir.strpath, request.param)
    return path


@pytest.fixture
def alchemy_config(request, base_config):
    """Provide a config that is suitable for sqlalchemy stores."""
    config = copy.deepcopy(base_config)
    # MEH/2020-01-07: (lb): This changes nothing; already the base_config default...
    #                       was that the original intent?
    config['db'].update({
        'orm': 'sqlalchemy',
        'engine': 'sqlite',
        'path': ':memory:',
    })
    return config


@pytest.fixture(
    params=(
        # SQLite
        {
            'engine': 'sqlite',
            'path': ':memory:',
        },
        # Non-SQLite
        {
            'engine': 'postgres',
            'host': fauxfactory.gen_ipaddr(),
            'port': fauxfactory.gen_integer(),
            'name': fauxfactory.gen_utf8(),
            'user': fauxfactory.gen_utf8(),
            'password': fauxfactory.gen_utf8(),
        },
        {
            'engine': 'postgres',
            'host': fauxfactory.gen_ipaddr(),
            'name': fauxfactory.gen_utf8(),
            'user': fauxfactory.gen_utf8(),
            'password': fauxfactory.gen_utf8(),
        },
    ),
)
def alchemy_config_parametrized(request, alchemy_config):
    """
    Provide a parametrized config that is suitable for sqlalchemy stores.

    We need to build our expectation dynamically if we want to use faked config values.
    """
    config = copy.deepcopy(alchemy_config)
    config['db'].update(request.param)
    if request.param['engine'] == 'sqlite':
        expectation = '{engine}:///{path}'.format(
            engine=request.param['engine'], path=request.param['path'])
    else:
        port = request.param.get('port', '')
        if port:
            port = ':{}'.format(port)
        expectation = '{engine}://{user}:{password}@{host}{port}/{name}'.format(
            engine=request.param['engine'],
            user=request.param['user'],
            password=request.param['password'],
            host=request.param['host'],
            port=port,
            name=request.param['name'],
        )
    return (config, expectation)


@pytest.fixture(params=(
    {'engine': None,
     'path': None},
    # sqlite
    {'engine': 'sqlite',
     'path': None},
    {'engine': 'sqlite',
     'path': ''},
    # Non-sqlite
    {'engine': 'postgres',
     'host': None,
     'name': fauxfactory.gen_utf8(),
     'user': fauxfactory.gen_utf8(),
     'password': fauxfactory.gen_alphanumeric()},
    {'engine': 'postgres',
     'host': fauxfactory.gen_ipaddr(),
     'name': '',
     'user': fauxfactory.gen_utf8(),
     'password': fauxfactory.gen_alphanumeric()},
    {'engine': 'postgres',
     'host': fauxfactory.gen_ipaddr(),
     'name': fauxfactory.gen_utf8(),
     'user': '',
     'password': fauxfactory.gen_alphanumeric()},
    {'engine': 'postgres',
     'host': fauxfactory.gen_ipaddr(),
     'name': fauxfactory.gen_utf8(),
     'user': fauxfactory.gen_utf8(),
     'password': ''}
))
def alchemy_config_missing_store_config_parametrized(request, alchemy_config):
    """
    Provide an alchemy config containing invalid key/value pairs for
    store initialization.
    """
    config = copy.deepcopy(alchemy_config)
    config['db'].update(request.param)
    return config


@pytest.fixture
@patch('nark.backends.sqlalchemy.storage.create_engine',
       lambda *args, **kwargs: None)
@patch('nark.backends.sqlalchemy.objects.metadata.create_all',
       lambda *args, **kwargs: None)
@patch('sqlalchemy_migrate_hotoffthehamster.versioning.api.db_version',
       lambda *args, **kwargs: None)
@patch('sqlalchemy_migrate_hotoffthehamster.versioning.api.downgrade',
       lambda *args, **kwargs: None)
@patch('sqlalchemy_migrate_hotoffthehamster.versioning.api.upgrade',
       lambda *args, **kwargs: None)
@patch('sqlalchemy_migrate_hotoffthehamster.versioning.api.version_control',
       lambda *args, **kwargs: None)
# (lb): 'sqlalchemy_migrate_hotoffthehamster.versioning.api.version'
#       is more complicated; see with-patch, below.
def alchemy_store(request, alchemy_config, alchemy_runner, alchemy_session):
    """
    Provide a SQLAlchemyStore that uses our test-session.

    Note:
        The engine created as part of the store.__init__() goes simply unused.
    """
    # (lb): There was a note from hamster-lib:
    #   "probably want this to autouse=True"
    # but after reviewing the docs, I'm guessing there's no point, because an
    # autouse fixture does not return a value, and the code we'd be replacing
    # -- alchemy_session -- creates the common.Session() object that each test
    # uses. So if we removed alchemy_session from the parameters for each test,
    # each test would still have to find the session object somehow.
    #   https://docs.pytest.org/en/latest/fixture.html#autouse-fixtures-xunit-setup-on-steroids
    store = SQLAlchemyStore(alchemy_config)
    with patch(
        'sqlalchemy_migrate_hotoffthehamster.versioning.api.version',
        new_callable=PropertyMock,
    ) as mock_version:
        # return_value does not matter so long as it's an int.
        type(mock_version).value = PropertyMock(return_value=3)
        store.standup(alchemy_session)
    return store


@pytest.fixture
def alchemy_activity_deleted(alchemy_activity_factory):
    alchemy_activity = alchemy_activity_factory()
    alchemy_activity.deleted = True
    return alchemy_activity


# Instance sets
# Convenience fixtures that provide multitudes of certain alchemy instances.
@pytest.fixture
def set_of_categories(alchemy_category_factory):
    """Provide a number of perstent facts at once."""
    return [alchemy_category_factory() for i in range(5)]


@pytest.fixture
def set_of_tags(alchemy_tag_factory):
    """Provide a number of perstent facts at once."""
    return [alchemy_tag_factory() for i in range(5)]


@pytest.fixture
def set_of_alchemy_facts(start_datetime, alchemy_fact_factory):
    """
    Provide a multitude of generic persistent facts.

    Facts have one day offset from each other and last 20 minutes each.
    """
    start = start_datetime
    result = []
    for i in range(5):
        end = start + datetime.timedelta(minutes=20)
        fact = alchemy_fact_factory(start=start, end=end)
        result.append(fact)
        start = start + datetime.timedelta(days=1)
    return result


# Fallback nark object and factory fixtures. Unless we know how factories
# interact.
@pytest.fixture
def category_factory(request, name):
    """Provide a ``nark.Category`` factory."""
    def generate():
        return Category(name, None)
    return generate


@pytest.fixture
def category(request, category_factory):
    """Provide a randomized ``nark.Category`` instance."""
    return category_factory()


@pytest.fixture
def tag_factory(request, name):
    """Provide a ``nark.Tag`` factory."""
    def generate():
        return Tag(name, None)
    return generate


@pytest.fixture
def tag(request, tag_factory):
    """Provide a randomized ``nark.Tag`` instance."""
    return tag_factory()


@pytest.fixture
def activity_factory(request, name, category_factory):
    """
    Provide a ``nark.Activity`` factory.

    Note:
        * The returned activity will have a *new* category associated as well.
        * Values are randomized but *not parametrized*.
    """
    def generate():
        category = category_factory()
        return Activity(name, pk=None, category=category, deleted=False)
    return generate


@pytest.fixture
def activity(request, activity_factory):
    """Provide a randomized ``nark.Activity`` instance."""
    return activity_factory()


@pytest.fixture
def fact_factory(
    request, activity_factory, tag_factory, start_end_datetimes, description,
):
    """
    Provide a ``nark.Fact`` factory.

    Note:
        * The returned fact will have a *new* activity (and by consequence category)
          associated as well.
        * Values are randomized but *not parametrized*.
    """
    def generate():
        activity = activity_factory()
        tags = set([tag_factory() for i in range(1)])
        start, end = start_end_datetimes
        fact = Fact(activity, start, end, pk=None, description=description, tags=tags)
        return fact
    return generate


@pytest.fixture
def fact(request, fact_factory):
    """Return a randomized ``nark.Fact`` instance."""
    return fact_factory()




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""``nark`` storage object managers."""

from gettext import gettext as _

from sqlalchemy.exc import IntegrityError

__all__ = (
    'BaseAlchemyManager',
    'query_apply_limit_offset',
    'query_apply_true_or_not',
)


def query_apply_limit_offset(query, **kwargs):
    """
    Applies 'limit' and 'offset' to the database fetch query

    On applies 'limit' if specified; and only applies 'offset' if specified.

    Args:
        query (???): Query (e.g., return from self.store.session.query(...))

        kwargs (keyword arguments):
            limit (int|str, optional): Limit to apply to the query.

            offset (int|str, optional): Offset to apply to the query.

    Returns:
        list: The query passed in, modified with limit and/or offset, maybe.
    """
    try:
        if kwargs['limit']:
            query = query.limit(kwargs['limit'])
    except KeyError:
        pass
    try:
        if kwargs['offset']:
            query = query.offset(kwargs['offset'])
    except KeyError:
        pass
    return query


def query_apply_true_or_not(query, column, condition, **kwargs):
    if condition is not None:
        return query.filter(column == condition)
    return query


class BaseAlchemyManager(object):
    """Base class for sqlalchemy managers."""

    # ***

    def add_and_commit(self, alchemy_item, raw=False, skip_commit=False):
        """
        Adds the item to the datastore, and perhaps calls commit.

        Generally, unless importing Facts, the session is committed
        after an item is added or updated. However, when adding or
        updating a Fact, we might also create other items (activity,
        category, tags), so we delay committing until everything is
        added/updated.
        """
        def _add_and_commit():
            session_add()
            session_commit_maybe()
            result = prepare_item()
            self.store.logger.debug(_("Added item: {!r}".format(result)))
            return result

        def session_add():
            self.store.session.add(alchemy_item)

        def session_commit_maybe():
            if skip_commit:
                return
            try:
                self.store.session.commit()
            except IntegrityError as err:
                message = _(
                    "An error occured! Are you sure that the {0}'s name "
                    "or ID is not already present? Error: '{1}'.".format(
                        self.__class__.__name__, err,
                    )
                )
                self.store.logger.error(message)
                raise ValueError(message)

        def prepare_item():
            result = alchemy_item
            if not raw:
                result = alchemy_item.as_hamster(self.store)
            return result

        return _add_and_commit()




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Raw SQLite3 commands to upgrade legacy hamster-applet database."""

from gettext import gettext as _

import logging
import sqlite3

logger = logging.getLogger('nark.log')


# (lb): Because SQLite3 does not support ALTER TABLE ... DROP COLUMN,
# we instead rename all tables, recreate new tables with old names,
# and then shuffle everything over. Enjoy!

def upgrade_legacy_db_hamster_applet(db_path):
    """"""
    def _upgrade_legacy_db_hamster_applet():
        # The caller will have copied the legacy db to db_path.
        conn, curs = begin_transaction()
        verify_legacy_version(curs)
        rename_old_tables(curs)
        create_new_tables(curs)
        populate_tables(curs)
        drop_old_tables(curs)
        commit_transaction(conn)

    def begin_transaction():
        # (lb): sqlite3 commits automatically. Tell it not to.
        conn = sqlite3.connect(db_path, isolation_level=None)
        curs = conn.cursor()
        # (lb): Not sure BEGIN TRANSACTION does anything, because isolation_level.
        curs.execute('BEGIN TRANSACTION')
        return conn, curs

    def commit_transaction(conn):
        conn.commit()
        conn.close()

    def verify_legacy_version(curs):
        curs.execute('SELECT * FROM version')
        versions = curs.fetchall()
        if len(versions) > 1:
            raise Exception(_(
                'Found more than 1 version entry: {}'
            ).format(versions))
        elif not len(versions):
            raise Exception(_('That database has no version entry.'))
        else:
            db_version = versions[0][0]
            logger.debug('Legacy DB Version: {}'.format(db_version))
            if db_version != 9:
                raise Exception(_(
                    "ERROR: Expected Legacy DB Version “9”, but found: {}"
                ).format(db_version))

    table_names = ['activities', 'categories', 'tags', 'facts', 'fact_tags']

    def rename_old_tables(curs):
        for table in table_names:
            curs.execute(
                'ALTER TABLE {table} RENAME TO temp_{table}'
                .format(table=table)
            )

    def create_new_tables(curs):
        """"""
        # NOTE: (lb): To form the CREATE TABLE statements, I simply ran
        #  ``.schema`` against a new DB created by ``dob store create``.
        create_categories(curs)
        create_facts(curs)
        create_fact_tags(curs)
        create_tags(curs)
        create_activities(curs)

    def create_categories(curs):
        # NOTE: (lb): Not recreating these legacy columns from categories:
        #   color_code varchar2(50)  # empty ((lb): at least in my legacy db)
        #   category_order integer   # empty ((lb): at least in my legacy db)
        #   search_name varchar2     # simply, name.lower()
        curs.execute(
            '''
            CREATE TABLE categories (
                id INTEGER NOT NULL,
                name VARCHAR(254),
                PRIMARY KEY (id),
                UNIQUE (name)
            )
            '''
        )

    def create_activities(curs):
        # NOTE: (lb): Not recreating these legacy columns from activities:
        #   work integer            # empty ((lb): at least in my legacy db)
        #   activity_order integer  # empty ((lb): at least in my legacy db)
        #   search_name varchar2    # simply, name.lower()
        curs.execute(
            '''
            CREATE TABLE activities (
                id INTEGER NOT NULL,
                name VARCHAR(500),
                deleted BOOLEAN,
                category_id INTEGER,
                PRIMARY KEY (id),
                UNIQUE (name, category_id),
                CHECK (deleted IN (0, 1)),
                FOREIGN KEY(category_id) REFERENCES categories (id)
            )
            '''
        )

    def create_tags(curs):
        # NOTE: (lb): Not recreating these legacy columns from tags:
        #   autocomplete BOOL DEFAULT true
        #   # (lb): I think autocomplete was the list that the legacy
        #   # ``hamster-indicator`` app let you edit in the interface.
        #   # This option is reimplemented using new ``hidden`` columns.
        curs.execute(
            '''
            CREATE TABLE tags (
                id INTEGER NOT NULL,
                name VARCHAR(254),
                PRIMARY KEY (id),
                UNIQUE (name)
            )
            '''
        )

    def create_facts(curs):
        curs.execute(
            '''
            CREATE TABLE facts (
                id INTEGER NOT NULL,
                start_time DATETIME,
                end_time DATETIME,
                activity_id INTEGER,
                description VARCHAR(500),
                PRIMARY KEY (id),
                FOREIGN KEY(activity_id) REFERENCES activities (id)
            )
            '''
        )

    def create_fact_tags(curs):
        curs.execute(
            '''
            CREATE TABLE fact_tags (
                fact_id INTEGER,
                tag_id INTEGER,
                FOREIGN KEY(fact_id) REFERENCES facts (id),
                FOREIGN KEY(tag_id) REFERENCES tags (id)
            )
            '''
        )

    def populate_tables(curs):
        populate_activities(curs)
        populate_categories(curs)
        populate_tags(curs)
        populate_facts(curs)
        populate_fact_tags(curs)

    def populate_activities(curs):
        curs.execute(
            '''
            INSERT INTO activities
            SELECT
                id, name, deleted, category_id
            FROM
                temp_activities
            '''
        )

    def populate_categories(curs):
        curs.execute(
            '''
            INSERT INTO categories
            SELECT
                id, name
            FROM
                temp_categories
            '''
        )

    def populate_tags(curs):
        curs.execute(
            '''
            INSERT INTO tags
            SELECT
                id, name
            FROM
                temp_tags
            '''
        )

    def populate_facts(curs):
        curs.execute(
            '''
            INSERT INTO facts
            SELECT
                id, start_time, end_time, activity_id, description
            FROM
                temp_facts
            '''
        )

    def populate_fact_tags(curs):
        curs.execute(
            '''
            INSERT INTO fact_tags
            SELECT
                fact_id, tag_id
            FROM
                temp_fact_tags
            '''
        )

    def drop_old_tables(curs):
        drop_tmp_tables(curs)
        drop_legacy_goo(curs)

    def drop_tmp_tables(curs):
        for table in reversed(table_names):
            curs.execute(
                'DROP TABLE temp_{table}'
                .format(table=table)
            )

    def drop_legacy_goo(curs):
        # Other legacy stuff to drop.
        drop_table_version(curs)

    def drop_table_version(curs):
        # (lb): Had one row, with one cell (version integer), set to "9".
        #   CREATE TABLE version(version integer);
        curs.execute('DROP TABLE version')

    def drop_old_fact_index(curs):
        # (lb): I think this virtual table was for some sort of search feature.
        #   It runs functions.
        #
        # From hamster-applet/src/hamster/db.py:
        #
        #   def run_fixtures(self):
        #       ...
        #       if version < 9:
        #           # adding full text search
        #           self.execute("""
        #               CREATE VIRTUAL TABLE fact_index
        #               USING fts3(id, name, category, description, tag)""")
        #
        # See:
        #
        #   https://www.sqlite.org/fts3.html
        #
        # NOTE: Dropping fact_index automatically drops three related tables:
        #
        #   fact_index_content
        #   fact_index_segments
        #   fact_index_segdir
        #
        curs.execute('DROP TABLE fact_index')

    _upgrade_legacy_db_hamster_applet()




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

import gettext
import importlib

from .config import REGISTERED_BACKENDS, decorate_config
from .helpers import logging as logging_helpers
from .helpers.dev.profiling import timefunc

# (lb): hamster-lib comment:
#
#         See: https://wiki.python.org/moin/PortingToPy3k/BilingualQuickRef#gettext
#         [FIXME]
#         Is this correct?
#           http://www.wefearchange.org/2012/06/the-right-way-to-internationalize-your.html
#         seems to use ``sys.version_info.major > 3``
#
# 2020-01-16: Now that Python2 (and six and future) dropped,
#  this branch not executed.
#
# FIXME/2020-01-16: (lb): EXPLAIN: Is gettext.install still necessary?
#  I.e., was this call just in case kwargs['unicode'] got set?
#
#  - Here's the old (v2-compatible) code:
#
#       kwargs = {}
#       if sys.version_info.major < 3:
#           kwargs['unicode'] = True
#       gettext.install('nark', **kwargs)
#
gettext.install('nark')


class NarkControl(object):
    """
    All mandatory config options are set as part of the controller setup.
    Any client may overwrite those values. but we can always asume that the
    controller does have a value set.

    We will try hard to get through with at least always returning the object.
    We should be able to change only the internal service code to then
    decompose it into its required weired format.

    We were compelled to make attach CRUD-methods to our activity, category
    and fact objects. But as all of those depend on access to the store
    anyway they seem to be best be placed here as a central hub.

    Generic CRUD-actions is to be delegated to our store. The Controller itself
    provides general timetracking functions so that our clients do not have to.
    """

    def __init__(self, config=None):
        if config:
            self.capture_config_lib(config)

    def capture_config_lib(self, config):
        self.config = decorate_config(config)
        self.lib_logger = self._get_logger()
        # Profiling: _get_store(): Observed: ~ 0.136 to 0.240 secs.
        self.store = self._get_store()
        self.sql_logger = self._sql_logger()

    def standup_store(self):
        created_fresh = self.store.standup()
        return created_fresh

    @property
    def migrations(self):
        return self.store.migrations

    @property
    def categories(self):
        return self.store.categories

    @property
    def activities(self):
        return self.store.activities

    @property
    def tags(self):
        return self.store.tags

    @property
    def facts(self):
        return self.store.facts

    # 2020-01-07: (lb): Only called by 1 test.
    def update_config(self, config):
        """Use a new config dictionary and apply its settings."""
        self.config = decorate_config(config)
        self.store = self._get_store()

    @timefunc
    def _get_store(self):
        """
        Setup the store used by this controller.

        This method is in charge off figuring out the store type, its
        instantiation as well as all additional configuration.
        """

        backend = REGISTERED_BACKENDS.get(self.config['db.orm'])
        if not backend:
            raise KeyError(_("No or invalid storage specified."))
        import_path, storeclass = tuple(backend.store_class.rsplit('.', 1))
        # Profiling: importlib.import_module: ~ 0.265 secs.
        backend_module = importlib.import_module(import_path)
        # storeclass, typically 'SQLAlchemyStore'.
        cls = getattr(backend_module, storeclass)
        store = cls(self.config)
        return store

    def _get_logger(self):
        """
        Setup and configure the main logger.

        As the docs suggest we setup just a pseudo handler. Any client that
        actually wants to use logging needs to setup its required handlers
        itself.
        """

        lib_log_level = self.config['dev.lib_log_level']
        lib_logger = logging_helpers.set_logger_level(
            'nark.log', lib_log_level,
        )
        return lib_logger

    def _sql_logger(self):
        """
        Setup and configure the SQLAlchemy database store logger.

        As the docs suggest we setup just a pseudo handler. Any client that
        actually wants to use logging needs to setup its required handlers
        itself.
        """
        return self.store.logger




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""
This module provides the database layout.

We inherit from our nark objects in order to use the custom
methods, making instance comparisons so much easier.

The reason we are not mapping our native nark objects directly is
that this seems to break the flexible plugable backend architecture
as SQLAlchemy establishes the mapping right away. This may be
avoidable and should be investigates later on.

If those classes are instantiated manually any nested related
instance needs to be added manually.

Note:

    Our dedicated SQLAlchemy objects do not perform any general data
    validation as not to duplicate code. This is expected to be
    handled by the generic ``nark`` objects.

    If need for backend specific validation should arise, it could of
    cause be added here.
"""

from gettext import gettext as _

# Profiling: Loading sqlalchemy takes about ~ 0.150 secs.
# (lb): And there's probably not a way to avoid it.
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    MetaData,
    Table,
    Unicode,
    UnicodeText,
    UniqueConstraint
)
from sqlalchemy.orm import mapper, relationship

from ...items.activity import Activity
from ...items.category import Category
from ...items.fact import Fact
from ...items.tag import Tag

DEFAULT_STRING_LENGTH = 254


class AlchemyCategory(Category):
    def __init__(self, pk, name, deleted, hidden):
        """
        Initiate a new SQLAlchemy category instance.

        Raises:
            TypeError: If ``category`` is not a ``Category`` instance.
        """

        self.pk = pk
        self.name = name
        self.deleted = bool(deleted)
        self.hidden = bool(hidden)

    def as_hamster(self, store):
        """Return store object as a real ``nark.Category`` instance."""
        return Category(
            pk=self.pk,
            name=self.name,
            deleted=bool(self.deleted),
            hidden=bool(self.hidden),
        )


class AlchemyActivity(Activity):
    def __init__(self, pk, name, category, deleted, hidden):
        """
        Initiate a new instance.

        Args:
            activity (nark.Activity): An activity that is to be
                represented as a backend instance.

        Raises:
            TypeError: If ``activity`` is not an ``Activity`` instance.
        """

        self.pk = pk
        self.name = name
        self.category = category
        self.deleted = bool(deleted)
        self.hidden = bool(hidden)

    def as_hamster(self, store):
        """Return new ``nark.Activity`` representation of SQLAlchemy instance."""
        if self.category:
            category = self.category.as_hamster(store)
        else:
            category = None
        activity_name = self.name
        # Play nice with "corrupt" databases (lacking integrity):
        #   Allow nameless items, rather than raising ValueError.
        # User can easily discover problem through CLI, and they can fix it.
        # Try this to find nameless activities and their IDs:
        #   dob list activities ''
        # Then, e.g.,
        #   dob edit activity #24 newname@newcategory
        if not activity_name:
            if self.deleted:
                activity_name = '<unnamed>'
                store.logger.warn(_('Activity in database has no name: {}').format(self))
            # else, let Activity() raise ValueError; at least for now.
            # MAYBE: else, if this is an issue, add a migration option to
            #              fix nameless activities. Or something.
        return Activity(
            pk=self.pk,
            name=activity_name,
            category=category,
            deleted=bool(self.deleted),
            hidden=bool(self.hidden),
        )


class AlchemyTag(Tag):
    def __init__(self, pk, name, deleted, hidden):
        """
        Initiate a new SQLAlchemy tag instance.

        Raises:
            TypeError: If ``category`` is not a ``Category`` instance.
        """

        self.pk = pk
        self.name = name
        self.deleted = bool(deleted)
        self.hidden = bool(hidden)

    def as_hamster(self, store):
        """Provide an convenient way to return it as a ``nark.Tag`` instance."""
        return Tag(
            pk=self.pk,
            name=self.name,
            deleted=bool(self.deleted),
            hidden=bool(self.hidden),
        )


class AlchemyFact(Fact):
    def __init__(self, pk, activity, start, end, description, deleted, split_from):
        """
        Initiate a new instance.

        Args:
            fact (nark.Fact): A fact that is to be represented
                as a backend instance.

        Raises:
            TypeError: If ``fact`` is not an ``Fact`` instance.
        """
        # FIXME/2018-05-15: (lb): DRY: Any reason this doesn't called super()?
        self.pk = pk
        self.deleted = bool(deleted)
        self.split_from = split_from
        self.activity = activity
        self.start = start
        self.end = end
        self.description = description
        # Tags can only be assigned after the fact has been created.
        self.tags = list()

    def as_hamster(self, store, tags=None):
        """Provide an convenient way to return it as a ``nark.Fact`` instance."""
        # NOTE: (lb): By default, self.tags is lazy loaded, which causes a fetch
        #   when it's looked up, once per Fact. This is normally not an issue, but
        #   I noticed a significant delay processing 15K Facts. My first attempt
        #   to resolve this was explicitly joining and coalescing tags in the
        #   query, and then passing the hydrated facts herein. So I added this
        #   if-else block. 2018-06-28: But then I learned about joinedload. I'm
        #   going to leave this here to make it easy to test performance issues
        #   as I continue to investigate this issue.
        if tags is None:
            nark_tags = set([tag.as_hamster(store) for tag in self.tags])
        else:
            nark_tags = set([Tag(tag) for tag in tags])

        fact_cls = store.fact_cls or Fact

        return fact_cls(
            pk=self.pk,
            deleted=bool(self.deleted),
            split_from=self.split_from,
            activity=self.activity.as_hamster(store),
            start=self.start,
            end=self.end,
            description=self.description,
            tags=nark_tags,
        )


metadata = MetaData()

categories = Table(
    'categories', metadata,
    Column('id', Integer, primary_key=True),

    # FIXME/2018-05-20: (lb): Why the hard limit? And why isn't it documented?
    Column('name', Unicode(DEFAULT_STRING_LENGTH), unique=True),

    Column('deleted', Boolean),
    Column('hidden', Boolean),
)

# (lb): This code uses SQLAlchemy Classical Mappings, and not Declarative Mappings.
#   http://docs.sqlalchemy.org/en/latest/orm/mapping_styles.html

mapper(AlchemyCategory, categories, properties={
    'pk': categories.c.id,
})

activities = Table(
    'activities', metadata,
    Column('id', Integer, primary_key=True),

    # FIXME/2018-05-20: (lb): Why the hard limit? And why isn't it documented?
    # And why isn't this DEFAULT_STRING_LENGTH instead of 500?
    Column('name', Unicode(500)),

    Column('deleted', Boolean),
    Column('hidden', Boolean),
    Column('category_id', Integer, ForeignKey(categories.c.id)),
    UniqueConstraint('name', 'category_id'),
)

mapper(AlchemyActivity, activities, properties={
    'pk': activities.c.id,
    'category': relationship(AlchemyCategory, backref='activities'),
})

tags = Table(
    'tags', metadata,
    Column('id', Integer, primary_key=True),

    # FIXME/2018-05-20: (lb): Why the hard limit? And why isn't it documented?
    Column('name', Unicode(DEFAULT_STRING_LENGTH), unique=True),

    Column('deleted', Boolean),
    Column('hidden', Boolean),
)

mapper(AlchemyTag, tags, properties={
    'pk': tags.c.id,
})

facts = Table(
    'facts', metadata,
    Column('id', Integer, primary_key=True),
    Column('deleted', Boolean),
    Column('split_from_id', Integer, ForeignKey('facts.id'), nullable=True),
    # SKIP: Column('hidden', Boolean),
    # NOTE/2018-04-22: Old Timey Hamster uses SQLite 'timestamp' data type.
    # In ProjectHamster Hamster, the data type shows as DATETIME. The type
    # is more of a suggestion in SQLite, which stores both types as strings,
    # and the strings are your typical datetime (iso8601 without the 'T',
    # and with a timezone), "YYYY-MM-DD HH:MM:SS".
    Column('start_time', DateTime),
    Column('end_time', DateTime),
    Column('activity_id', Integer, ForeignKey(activities.c.id)),

    # FIXME/2018-05-20: (lb): Why the hard limit? And why isn't it documented?
    # ALSO: seriously, only 500 chars for "description"??
    #    Column('description', Unicode(500)),
    # [In SQLite the size is meaningless (you can store anything, any size)
    #  but it probably matters in other databases]...
    # FIXME/2018-06-09: (lb): Remove this comment after verifying against
    #   another store, e.g., Postgres.
    Column('description', UnicodeText()),
)

mapper(AlchemyFact, facts, properties={
    'pk': facts.c.id,
    'activity': relationship(AlchemyActivity, backref='facts'),
    'tags': relationship(AlchemyTag, backref='facts', secondary=lambda: fact_tags),

    # 2018-04-22: (lb): I'm not sure if there's a migration script or not,
    # but I could not find one, and for some reason ProjectHamster renamed
    # the facts' start and end columns to start_time and end_time, respectively.
    # FIXME: (lb): Remove this comment (and newlines) after verifying it's eh-okay.
    'start': facts.c.start_time,
    'end': facts.c.end_time,

    'split_from': relationship(
        lambda: AlchemyFact, remote_side=facts.c.id, backref='sub_facts',
    )
})

# 2018-04-22: (lb): ProjectHamster renamed fact_tags to facttags. But
# that term isn't used in the code other than in this Table mapping
# (which no other code uses; though maybe SQLAlchemy uses it internally?).
fact_tags = Table(
    'fact_tags', metadata,
    Column('fact_id', Integer, ForeignKey(facts.c.id)),
    Column('tag_id', Integer, ForeignKey(tags.c.id)),
)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Fixtures that are of general use."""

import datetime

import faker as faker_
import pytest
from pytest_factoryboy import register

from nark.control import NarkControl
from nark.manager import BaseStore
from nark.tests import factories

register(factories.CategoryFactory)
register(factories.ActivityFactory)
register(factories.TagFactory)
register(factories.FactFactory)

faker = faker_.Faker()


def convert_time_to_datetime(time_string):
    """
    Helper method.

    If given a %H:%M string, return a datetime.datetime object with todays
    date.
    """
    return datetime.datetime.combine(
        # MAYBE: Use controller.store.now ?
        datetime.datetime.utcnow().date(),
        datetime.datetime.strptime(time_string, "%H:%M").time()
    )


# Controller


@pytest.yield_fixture
def controller(base_config):
    """Provide a basic controller."""
    # [TODO] Parametrize over all available stores.
    controller = NarkControl(base_config)
    yield controller
    controller.store.cleanup()


@pytest.fixture
def basestore(base_config):
    """Provide a generic ``storage.BaseStore`` instance using ``baseconfig``."""
    store = BaseStore(base_config)
    return store


# Categories
@pytest.fixture(params=(None, True,))
def category_valid_parametrized(
    request, category_factory, name_string_valid_parametrized,
):
    """Provide a variety of valid category fixtures."""
    if request.param:
        result = category_factory(name=name_string_valid_parametrized)
    else:
        result = None
    return result


@pytest.fixture
def category_valid_parametrized_without_none(
    request, category_factory, name_string_valid_parametrized,
):
    """
    Provide a parametrized category fixture but not ``None``.

    This fixuture will represent a wide array of potential name charsets as well
    but not ``category=None``.
    """
    return category_factory(name=name_string_valid_parametrized)


# Activities
@pytest.fixture
def activity_valid_parametrized(
    request,
    activity_factory,
    name_string_valid_parametrized,
    category_valid_parametrized,
    deleted_valid_parametrized,
):
    """Provide a huge array of possible activity versions. Including None."""
    return activity_factory(
        name=name_string_valid_parametrized,
        category=category_valid_parametrized,
        deleted=deleted_valid_parametrized,
    )


@pytest.fixture
def new_activity_values(category):
    """Return garanteed modified values for a given activity."""
    def modify(activity):
        return {
            'name': activity.name + 'foobar',
        }
    return modify


# Facts
@pytest.fixture
def fact_factory():
    """Return a factory class that generates non-persisting Fact instances."""
    return factories.FactFactory.build


@pytest.fixture
def fact():
    """Provide a randomized non-persistant Fact-instance."""
    return factories.FactFactory.build()


@pytest.fixture
def list_of_facts(fact_factory):
    """
    Provide a factory that returns a list with given amount of Fact instances.

    The key point here is that these fact *do not overlap*!
    """
    def get_list_of_facts(number_of_facts):
        facts = []
        # MAYBE: Use controller.store.now ?
        old_start = datetime.datetime.utcnow().replace(microsecond=0)
        offset = datetime.timedelta(hours=4)
        for i in range(number_of_facts):
            start = old_start + offset
            facts.append(fact_factory(start=start))
            old_start = start
        return facts
    return get_list_of_facts


@pytest.fixture(params=('%M', '%H:%M', 'HHhMMm', ''))
def string_delta_style_parametrized(request):
    """Provide all possible format option for ``Fact().format_delta()``."""
    return request.param


@pytest.fixture
def today_fact(fact_factory):
    """Return a ``Fact`` instance that start and ends 'today'."""
    # MAYBE: Use controller.store.now ?
    start = datetime.datetime.utcnow()
    end = start + datetime.timedelta(minutes=30)
    return fact_factory(start=start, end=end)


@pytest.fixture
def not_today_fact(fact_factory):
    """Return a ``Fact`` instance that neither start nor ends 'today'."""
    # MAYBE: Use controller.store.now ?
    start = datetime.datetime.utcnow() - datetime.timedelta(days=2)
    end = start + datetime.timedelta(minutes=30)
    return fact_factory(start=start, end=end)


@pytest.fixture
def current_fact(fact_factory):
    """Provide a ``ongoing fact``. That is a fact that has started but not ended yet."""
    # MAYBE: Use controller.store.now ?
    return fact_factory(
        start=datetime.datetime.utcnow().replace(microsecond=0),
        end=None,
    )


@pytest.fixture(params=[
    '',
    '14:00 - 12:00 foo@bar',
    '12:00 - 14:00 @bar',
    '12:00:11 - 11:00:59 foo@bar',
])
def invalid_raw_fact_parametrized(request):
    """Return various invalid ``raw fact`` strings."""
    return request.param


@pytest.fixture
def raw_fact_with_persistent_activity(persistent_activity):
    """A raw fact whichs 'activity' is already present in the db."""
    return (
        '12:00-14:14 {a.name}@{a.category.name}'.format(a=persistent_activity), {
            'start': convert_time_to_datetime('12:00'),
            'end': convert_time_to_datetime('14:14'),
            'activity': persistent_activity.name,
            'category': persistent_activity.category.name,
            'description': None,
        },
    )




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Global fixtures."""

import datetime

import fauxfactory
import pytest
from pytest_factoryboy import register

from nark.config import decorate_config
from nark.tests import factories

# Make factory fixtures, like alchemy_category_factory.
#   ((lb) At least this is where I think it happens.)
register(factories.CategoryFactory)
register(factories.ActivityFactory)
register(factories.TagFactory)
register(factories.FactFactory)


# This fixture is used by ``test_helpers`` and ``test_storage``.
@pytest.fixture
def endless_fact(fact_factory):
    """Provide an existing 'ongoing fact'."""
    # (lb): Comment from hamster-lib:
    #   For reasons unknown ``fact.tags`` would be empty
    #   when using the ``fact`` fixture.
    fact = fact_factory()
    fact.end = None
    return fact


@pytest.fixture
def base_config(tmpdir):
    """Provide a generic baseline configuration."""
    base_config = {
        'db': {
            'orm': 'sqlalchemy',
            'engine': 'sqlite',
            'path': ':memory:',
        },
        'dev': {
            'lib_log_level': 'WARNING',
            'sql_log_level': 'WARNING',
        },
        'time': {
            # FIXME: (lb): Make special tests for these less used options
            #        and then just set to default values here, e.g.,
            #           'day_start': '',
            #           'fact_min_delta': 0,
            'day_start': datetime.time(hour=5, minute=30, second=0),
            'fact_min_delta': 60,
        },
    }
    # (lb): The application deals with a ConfigDecorator object, and not a
    # simple dict, which has the advantage that our tests (and any client
    # code) does not need to ensure that it sets all the config values.
    # - However, we still return a dictionary, because we want to be able
    # to change config values without triggering any validation. E.g.,
    # calling config_root['db'].update({'engine': None}) would raise.
    # But we don't want to blow up in the fixture!
    config_root = decorate_config(base_config)
    config = config_root.as_dict()
    return config


@pytest.fixture
def start_end_datetimes_from_offset_now():
    """Generate start/end datetime tuple with given offset in minutes."""
    def generate(offset):
        # MAYBE: Use controller.store.now?
        end = datetime.datetime.now().replace(microsecond=0)
        start = end - datetime.timedelta(minutes=offset)
        return (start, end)
    return generate


@pytest.fixture
# (lb): If shouldn't matter if we use now() or utcnow(). Right?
def start_end_datetimes_from_offset_utcnow():
    """Generate start/end datetime tuple with given offset in minutes."""
    def generate(offset):
        # MAYBE: Use controller.store.now?
        end = datetime.datetime.utcnow().replace(microsecond=0)
        start = end - datetime.timedelta(minutes=offset)
        return (start, end)
    return generate


@pytest.fixture(params=(True, False))
def bool_value_parametrized(request):
    """
    Return a parametrized boolean value.

    This is usefull to easily parametrize tests using flags.
    """
    return request.param


# Attribute fixtures (non-parametrized)
@pytest.fixture
def name():
    """Randomized, valid but non-parametrized name string."""
    return fauxfactory.gen_utf8()


@pytest.fixture
def start_end_datetimes(start_end_datetimes_from_offset_now):
    """Return a start/end-datetime-tuple."""
    return start_end_datetimes_from_offset_now(15)


@pytest.fixture
def start_datetime():
    """Provide an arbitrary datetime."""
    # (lb): Because Freezegun, datetime.now() is datetime.utcnow().
    return datetime.datetime.utcnow().replace(microsecond=0)


@pytest.fixture
def description():
    """Return a generic text suitable to mimic a ``Fact.description``."""
    return fauxfactory.gen_iplum()


# New value generation
@pytest.fixture
def new_category_values():
    """Return garanteed modified values for a given category."""
    def modify(category):
        return {
            'name': category.name + 'foobar',
        }
    return modify


@pytest.fixture
def new_tag_values():
    """Return garanteed modified values for a given tag."""
    def modify(tag):
        return {
            'name': tag.name + 'foobar',
        }
    return modify


@pytest.fixture
def new_fact_values(tag_factory, activity_factory):
    """Provide guaranteed different Fact-values for a given Fact-instance."""
    def modify(fact):
        if fact.end:
            end = fact.end - datetime.timedelta(days=10)
        else:
            end = None
        return {
            'activity': activity_factory(),
            'start': fact.start - datetime.timedelta(days=10),
            'end': end,
            'description': fact.description + 'foobar',
            'tags': set([tag_factory() for i in range(5)])
        }
    return modify


# Valid attributes parametrized
@pytest.fixture(params=('', 'cyrillic', 'utf8', ))
def name_string_valid_parametrized(request):
    """Provide a variety of strings that should be valid non-tag *names*."""
    if not request.param:
        return request.param
    return fauxfactory.gen_string(request.param)


@pytest.fixture(params=('cyrillic', 'utf8',))
def name_string_valid_parametrized_tag(request):
    """Provide a variety of strings that should be valid tag *names*."""
    return fauxfactory.gen_string(request.param)


@pytest.fixture(params=(None,))
def name_string_invalid_parametrized(request):
    """Provide a variety of strings that should be valid non-tag *names*."""
    return request.param


@pytest.fixture(params=(None, ''))
def name_string_invalid_parametrized_tag(request):
    """Provide a variety of strings that should be valid tag *names*."""
    return request.param


@pytest.fixture(params=(
    fauxfactory.gen_string('numeric'),
    fauxfactory.gen_string('alphanumeric'),
    fauxfactory.gen_string('utf8'),
    None,
))
def pk_valid_parametrized(request):
    """Provide a variety of valid primary keys.

    Note:
        At our current stage we do *not* make assumptions about the type of primary key.
        Of cause, this may be a different thing on the backend level!
    """
    return request.param


@pytest.fixture(params=(True, False, 0, 1, '', 'foobar'))
def deleted_valid_parametrized(request):
    """Return various valid values for the ``deleted`` argument."""
    return request.param


@pytest.fixture(params='alpha cyrillic latin1 utf8'.split())
def description_valid_parametrized(request):
    """Provide a variety of strings that should be valid *descriptions*."""
    return fauxfactory.gen_string(request.param)


@pytest.fixture(params='alpha cyrillic latin1 utf8'.split())
def tag_list_valid_parametrized(request):
    """Provide a variety of strings that should be valid *descriptions*."""
    return set([fauxfactory.gen_string(request.param) for i in range(4)])




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

""""""

import time
from functools import update_wrapper

from ... import __PROFILING__, __time_0__

__all__ = (
    'profile_elapsed',
    'timefunc',
    'timefunct',
    'TimeWith',
)


MSGS_FUNC = []
MSGS_SPAN = []


def capture_func(msg):
    MSGS_FUNC.append(msg)


def capture_span(msg):
    MSGS_SPAN.append(msg)


def profile_elapsed(text):
    if not __PROFILING__:
        return
    capture_span('{0}: {1:.3f} secs.'.format(text, time.time() - __time_0__))


# Thanks! The following is an edited version of:
#
#   https://zapier.com/engineering/profiling-python-boss/

def timefunc(func):
    if not __PROFILING__:
        return func

    def f_timer(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        capture_func('{0}: {1:.3f} secs.'.format(func.__name__, end - start))
        return result

    return f_timer


def timefunct(text=None):
    def _timefunc(func):
        if not __PROFILING__:
            return func

        def f_timer(*args, **kwargs):
            name = text if text is not None else func.__name__
            start = time.time()
            result = func(*args, **kwargs)
            end = time.time()
            capture_func('{0}: {1:.3f} secs.'.format(name, end - start))
            return result

        return update_wrapper(f_timer, func)
    return _timefunc


class TimeWith():
    """"""
    def __init__(self, name='', start=None):
        self.name = name
        self.start = time.time() if start is None else start

    @property
    def elapsed(self):
        return time.time() - self.start

    def checkpoint(self, name=''):
        capture_span('{timer} {checkpoint}: {elapsed} secs.'.format(
            timer=self.name,
            checkpoint=name,
            elapsed=self.elapsed,
        ).strip())

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.checkpoint('finished')
        pass


if __PROFILING__:
    def exit_elapsed():
        profile_elapsed('To dob:   exit')
        for msg in MSGS_SPAN:
            # T001 print found.
            print(msg)  # noqa: T001
        for msg in MSGS_FUNC:
            print(msg)  # noqa: T001

    import atexit
    atexit.register(exit_elapsed)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import logging
import sys

import pytest

from unittest.mock import patch

from nark.config.log_levels import get_log_level_safe, must_verify_log_level


class TestConfigLogLevels:
    def test_must_verify_log_level_valid(self, log_level_valid_parametrized):
        _log_level = must_verify_log_level(log_level_valid_parametrized)
        assert(isinstance(_log_level, int))

    def test_must_verify_log_level_invalid(self, log_level_invalid_parametrized):
        with pytest.raises(ValueError):
            must_verify_log_level(log_level_invalid_parametrized)

    def test_get_log_level_safe_valid(self, log_level_valid_parametrized):
        _log_level = get_log_level_safe(log_level_valid_parametrized)
        assert(isinstance(_log_level, int))

    def test_get_log_level_safe_invalid(self, log_level_invalid_parametrized):
        _log_level = get_log_level_safe(log_level_invalid_parametrized)
        assert(isinstance(_log_level, int))

    def test_get_log_level_on_completion(self):
        testargs = [
            "/home/user/.virtualenvs/dob/lib/python3.[5-7]/site-packages/pytest.py",
            "complete",
        ]
        with patch.object(sys, 'argv', testargs):
            _log_level = get_log_level_safe(logging.DEBUG)
            assert(_log_level == (logging.CRITICAL + 1))




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import datetime

import pytest
from freezegun import freeze_time


# ***

class TestBaseStore():
    def test_cleanup(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.cleanup()


# ***

class TestCategoryManager():
    def test_add(self, basestore, category):
        with pytest.raises(NotImplementedError):
            basestore.categories._add(category)

    def test_update(self, basestore, category):
        with pytest.raises(NotImplementedError):
            basestore.categories._update(category)

    def test_remove(self, basestore, category):
        with pytest.raises(NotImplementedError):
            basestore.categories.remove(category)

    def test_get_invalid_pk(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.categories.get(12)

    def test_get_invalid_pk_type(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.categories.get_by_name('fooo')

    def test_save_wrong_type(self, basestore, category):
        with pytest.raises(TypeError):
            basestore.categories.save([])

    def test_save_new(self, basestore, category, mocker):
        """Make sure that saving an new category calls ``__add``."""
        basestore.categories._add = mocker.MagicMock(return_value=category)
        try:
            basestore.categories.save(category)
        except NotImplementedError:
            pass
        assert basestore.categories._add.called

    def test_save_existing(self, basestore, category, mocker):
        category.pk = 0
        basestore.categories._update = mocker.MagicMock(return_value=category)
        try:
            basestore.categories.save(category)
        except NotImplementedError:
            pass
        assert basestore.categories._update.called

    def test_get_or_create_existing(self, basestore, category, mocker):
        """Make sure the category is beeing looked up and no new one is created."""
        basestore.categories.get_by_name = mocker.MagicMock(return_value=category)
        basestore.categories._add = mocker.MagicMock(return_value=category)
        try:
            basestore.categories.get_or_create(category.name)
        except NotImplementedError:
            pass
        assert basestore.categories._add.called is False
        assert basestore.categories.get_by_name.called

    def test_get_or_create_new_category(self, basestore, category, mocker):
        """Make sure the category is beeing looked up and new one is created."""
        basestore.categories._add = mocker.MagicMock(return_value=category)
        basestore.categories.get_by_name = mocker.MagicMock(side_effect=KeyError)
        try:
            basestore.categories.get_or_create(category.name)
        except NotImplementedError:
            pass
        assert basestore.categories.get_by_name.called
        assert basestore.categories._add.called

    def test_get_all(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.categories.get_all()


# ***

class TestActivityManager:
    def test_save_new(self, basestore, activity, mocker):
        """Make sure that saving an new activity calls ``_add``."""
        basestore.activities._add = mocker.MagicMock(return_value=activity)
        try:
            basestore.activities.save(activity)
        except NotImplementedError:
            pass
        assert basestore.activities._add.called

    def test_save_existing(self, basestore, activity, mocker):
        """Make sure that saving an existing activity calls ``_update``."""
        activity.pk = 0
        basestore.activities._update = mocker.MagicMock(return_value=activity)
        try:
            basestore.activities.save(activity)
        except NotImplementedError:
            pass
        assert basestore.activities._update.called

    def test_get_or_create_existing(self, basestore, activity, mocker):
        basestore.activities.get_by_composite = mocker.MagicMock(return_value=activity)
        basestore.activities.save = mocker.MagicMock(return_value=activity)
        result = basestore.activities.get_or_create(activity)
        assert result.name == activity.name
        assert basestore.activities.save.called is False

    def test_get_or_create_new(self, basestore, activity, mocker):
        basestore.activities.get_by_composite = mocker.MagicMock(side_effect=KeyError())
        basestore.activities.save = mocker.MagicMock(return_value=activity)
        result = basestore.activities.get_or_create(activity)
        assert result.name == activity.name
        assert basestore.activities.save.called is True

    def test_add(self, basestore, activity):
        with pytest.raises(NotImplementedError):
            basestore.activities._add(activity)

    def test_update(self, basestore, activity):
        with pytest.raises(NotImplementedError):
            basestore.activities._update(activity)

    def test_remove(self, basestore, activity):
        with pytest.raises(NotImplementedError):
            basestore.activities.remove(activity)

    def test_get_by_composite(self, basestore, activity):
        with pytest.raises(NotImplementedError):
            basestore.activities.get_by_composite(activity.name, activity.category)

    def test_get(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.activities.get(12)

    def test_get_all(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.activities.get_all()


# ***

class TestTagManager():
    def test_add(self, basestore, tag):
        with pytest.raises(NotImplementedError):
            basestore.tags._add(tag)

    def test_update(self, basestore, tag):
        with pytest.raises(NotImplementedError):
            basestore.tags._update(tag)

    def test_remove(self, basestore, tag):
        with pytest.raises(NotImplementedError):
            basestore.tags.remove(tag)

    def test_get_invalid_pk(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.tags.get(12)

    def test_get_invalid_pk_type(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.tags.get_by_name('fooo')

    def test_save_wrong_type(self, basestore, tag):
        with pytest.raises(TypeError):
            basestore.tags.save([])

    def test_save_new(self, basestore, tag, mocker):
        """Make sure that saving an new tag calls ``__add``."""
        basestore.tags._add = mocker.MagicMock(return_value=tag)
        try:
            basestore.tags.save(tag)
        except NotImplementedError:
            pass
        assert basestore.tags._add.called

    def test_save_existing(self, basestore, tag, mocker):
        tag.pk = 0
        basestore.tags._update = mocker.MagicMock(return_value=tag)
        try:
            basestore.tags.save(tag)
        except NotImplementedError:
            pass
        assert basestore.tags._update.called

    def test_get_or_create_existing(self, basestore, tag, mocker):
        """Make sure the tag is beeing looked up and no new one is created."""
        basestore.tags.get_by_name = mocker.MagicMock(return_value=tag)
        basestore.tags._add = mocker.MagicMock(return_value=tag)
        try:
            basestore.tags.get_or_create(tag.name)
        except NotImplementedError:
            pass
        assert basestore.tags._add.called is False
        assert basestore.tags.get_by_name.called

    def test_get_or_create_new_tag(self, basestore, tag, mocker):
        """Make sure the tag is beeing looked up and new one is created."""
        basestore.tags._add = mocker.MagicMock(return_value=tag)
        basestore.tags.get_by_name = mocker.MagicMock(side_effect=KeyError)
        try:
            basestore.tags.get_or_create(tag.name)
        except NotImplementedError:
            pass
        assert basestore.tags.get_by_name.called
        assert basestore.tags._add.called

    def test_get_all(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.tags.get_all()


# ***

class TestFactManager:
    def test_save_endless_fact(self, basestore, fact, mocker):
        """
        Make sure that passing a fact without end (aka 'endless, ongoing fact')
        triggers the right method.
        """
        magic_fact = {}
        basestore.facts._add = mocker.MagicMock(return_value=magic_fact)
        fact.end = None
        new_fact = basestore.facts.save(fact)
        assert basestore.facts._add.called
        assert new_fact is magic_fact

    def test_save_to_brief_fact(self, basestore, fact):
        """Ensure that a fact with too small of a time delta raises an exception."""
        delta = datetime.timedelta(seconds=(basestore.config['time.fact_min_delta'] - 1))
        fact.end = fact.start + delta
        with pytest.raises(ValueError):
            basestore.facts.save(fact)

    def test_add(self, basestore, fact):
        with pytest.raises(NotImplementedError):
            basestore.facts._add(fact)

    def test_update(self, basestore, fact):
        with pytest.raises(NotImplementedError):
            basestore.facts._update(fact)

    def test_remove(self, basestore, fact):
        with pytest.raises(NotImplementedError):
            basestore.facts.remove(fact)

    def test_get(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.facts.get(12)

    @pytest.mark.parametrize(('since', 'until', 'filter_term', 'expectation'), [
        (None, None, '', {
            'since': None,
            'until': None}),
        # Various since info.
        (datetime.date(2014, 4, 1), None, '', {
            'since': datetime.datetime(2014, 4, 1, 5, 30, 0),
            'until': None}),
        (datetime.time(13, 40, 25), None, '', {
            'since': datetime.datetime(2015, 4, 1, 13, 40, 25),
            'until': None}),
        (datetime.datetime(2014, 4, 1, 13, 40, 25), None, '', {
            'since': datetime.datetime(2014, 4, 1, 13, 40, 25),
            'until': None}),
        # Various until info.
        (None, datetime.date(2014, 2, 1), '', {
            'since': None,
            'until': datetime.datetime(2014, 2, 2, 5, 29, 59)}),
        (None, datetime.time(13, 40, 25), '', {
            'since': None,
            'until': datetime.datetime(2015, 4, 1, 13, 40, 25)}),
        (None, datetime.datetime(2014, 4, 1, 13, 40, 25), '', {
            'since': None,
            'until': datetime.datetime(2014, 4, 1, 13, 40, 25)}),
    ])
    @freeze_time('2015-04-01 18:00')
    def test_get_all_various_since_and_until_times(
        self, basestore, mocker, since, until, filter_term, expectation,
    ):
        """Test that time conversion matches expectations."""
        basestore.facts._get_all = mocker.MagicMock()
        basestore.facts.get_all(since=since, until=until, search_term=filter_term)
        assert basestore.facts._get_all.called
        assert basestore.facts._get_all.call_args[1] == {
            'since': expectation['since'],
            'until': expectation['until'],
            'search_term': filter_term,
        }

    @pytest.mark.parametrize(
        ('since', 'until'),
        [
            (
                datetime.date(2015, 4, 5),
                datetime.date(2012, 3, 4),
            ),
            (
                datetime.datetime(2015, 4, 5, 18, 0, 0),
                datetime.datetime(2012, 3, 4, 19, 0, 0),
            ),
        ],
    )
    def test_get_all_until_before_since(self, basestore, mocker, since, until):
        """Test that we throw an exception if passed until time is before since time."""
        with pytest.raises(ValueError):
            basestore.facts.get_all(since, until)

    @pytest.mark.parametrize(('since', 'until'), [
        (datetime.date(2015, 4, 5), '2012-03-04'),
        ('2015-04-05 18:00:00', '2012-03-04 19:00:00'),
    ])
    def test_get_all_invalid_date_types(self, basestore, mocker, since, until):
        """Test that we throw an exception if we recieve invalid date/time objects."""
        with pytest.raises(TypeError):
            basestore.facts.get_all(since, until)

    @freeze_time('2015-10-03 14:45')
    def test_get_today(self, basestore, mocker):
        """Make sure that method uses appropriate timeframe."""
        basestore.facts.get_all = mocker.MagicMock(return_value=[])
        result = basestore.facts.get_today()
        assert result == []
        assert (
            basestore.facts.get_all.call_args[1] == {
                'since': datetime.datetime(2015, 10, 3, 5, 30, 0),
                'until': datetime.datetime(2015, 10, 4, 5, 29, 59),
            }
        )

    def test__get_all(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.facts._get_all()

    @freeze_time('2019-02-01 18:00')
    @pytest.mark.parametrize('hint', (
        None,
        datetime.timedelta(minutes=10),
        datetime.timedelta(minutes=300),
        datetime.timedelta(seconds=-10),
        datetime.timedelta(minutes=-10),
        datetime.datetime(2019, 2, 1, 19),
    ))
    def test_stop_current_fact_with_hint(
        self, basestore, base_config, endless_fact, hint, mocker,
    ):
        """
        Make sure we can stop an 'ongoing fact' and that it will have an end set.

        Please note that ever so often it may happen that the factory generates
        a endless_fact with ``Fact.start`` after our mocked today-date. In order to avoid
        confusion the easies fix is to make sure the mock-today is well in the future.
        """
        now = datetime.datetime.now()  # the freeze_time time, above.
        if endless_fact.start > now:
            # (lb): The FactFactory sets start to faker.Faker().date_time(),
            # which is not constrained in any way (maybe it doesn't return
            # time from the future?). Here we dial back the start if it's
            # too far a’future, because then the FactManager will complain
            # about trying to end the Fact before it started.
            fmdelta = base_config['time']['fact_min_delta']
            endless_fact.start = now - datetime.timedelta(seconds=fmdelta)
        # NOTE: The `fact` fixture simply adds a second Fact to the db, after
        #       having added the endless_fact Fact.
        if hint:
            if isinstance(hint, datetime.datetime):
                expected_end = hint
            else:
                # (lb): This operation is same as = now + hint, isn't it.
                expected_end = datetime.datetime(2019, 2, 1, 18) + hint
        else:
            # NOTE: Because freeze_time, datetime.now() === datetime.utcnow().
            expected_end = datetime.datetime.now().replace(microsecond=0)
        basestore.facts.endless = mocker.MagicMock(return_value=[endless_fact])
        basestore.facts._add = mocker.MagicMock()
        basestore.facts.stop_current_fact(hint)
        assert basestore.facts.endless.called
        assert basestore.facts._add.called
        fact_to_be_added = basestore.facts._add.call_args[0][0]
        assert fact_to_be_added.end == expected_end
        fact_to_be_added.end = None
        assert fact_to_be_added == endless_fact

    @freeze_time('2019-02-01 18:00')
    @pytest.mark.parametrize('hint', (
        datetime.datetime(2019, 2, 1, 17, 59),
    ))
    def test_stop_current_fact_with_end_in_the_past(
        self, basestore, base_config, endless_fact, hint, mocker,
    ):
        """
        Make sure that stopping an 'ongoing fact' with end before start raises.
        """
        # Set start to the freeze_time time, above.
        endless_fact.start = datetime.datetime.now()
        basestore.facts.endless = mocker.MagicMock(return_value=[endless_fact])
        basestore.facts._add = mocker.MagicMock()
        with pytest.raises(ValueError):
            basestore.facts.stop_current_fact(hint)
        assert basestore.facts.endless.called
        assert not basestore.facts._add.called

    def test_stop_current_fact_invalid_offset_hint(
        self, basestore, endless_fact, mocker,
    ):
        """
        Make sure that stopping with an offset hint that results in end > start
        raises an error.
        """
        basestore.facts.endless = mocker.MagicMock(return_value=[endless_fact])
        now = datetime.datetime.now().replace(microsecond=0)
        offset = (now - endless_fact.start).total_seconds() + 100
        offset = datetime.timedelta(seconds=-1 * offset)
        with pytest.raises(ValueError):
            basestore.facts.stop_current_fact(offset)

    def test_stop_current_fact_invalid_datetime_hint(
        self, basestore, endless_fact, mocker,
    ):
        """
        Make sure that stopping with a datetime hint that results in end > start
        raises an error.
        """
        basestore.facts.endless = mocker.MagicMock(return_value=[endless_fact])
        with pytest.raises(ValueError):
            basestore.facts.stop_current_fact(
                endless_fact.start - datetime.timedelta(minutes=30),
            )

    def test_stop_current_fact_invalid_hint_type(
        self, basestore, endless_fact, mocker,
    ):
        """Make sure that passing an invalid hint type raises an error."""
        basestore.facts.endless = mocker.MagicMock(return_value=[endless_fact])
        with pytest.raises(TypeError):
            basestore.facts.stop_current_fact(str())

    def test_stop_current_fact_non_existing(self, basestore, mocker):
        """
        Make sure that trying to call stop when there is no 'endless fact'
        raises an error.
        """
        basestore.facts.endless = mocker.MagicMock(return_value=[])
        with pytest.raises(KeyError):
            basestore.facts.stop_current_fact()
            # KeyError: 'No ongoing Fact found.'

    def test_get_endless_fact_with_ongoing_fact(
        self, basestore, endless_fact, fact, mocker,
    ):
        """Make sure we return the 'ongoing_fact'."""
        basestore.facts.endless = mocker.MagicMock(return_value=[endless_fact])
        fact = basestore.facts.endless()
        assert fact == fact
        assert fact is fact

    def test_get_endless_fact_without_ongoing_fact(self, basestore, mocker):
        """Make sure that we raise a KeyError if ther is no 'ongoing fact'."""
        basestore.facts.endless = mocker.MagicMock(return_value=[])
        fact = basestore.facts.endless()
        assert fact == []

    def test_cancel_current_fact(self, basestore, endless_fact, fact, mocker):
        """Make sure we return the 'ongoing_fact'."""
        basestore.facts.endless = mocker.MagicMock(return_value=[endless_fact])
        basestore.facts.remove = mocker.MagicMock()
        result = basestore.facts.cancel_current_fact()
        assert basestore.facts.endless.called
        assert basestore.facts.remove.called
        assert result is endless_fact  # Because mocked.
        # FIXME: Where's the test that actually tests FactManager.endless()?

    def test_cancel_current_fact_without_endless_fact(self, basestore, mocker):
        """Make sure that we raise a KeyError if there is no 'ongoing fact'."""
        basestore.facts.endless = mocker.MagicMock(return_value=[])
        with pytest.raises(KeyError):
            basestore.facts.cancel_current_fact()
        assert basestore.facts.endless.called




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""This module provides several time-related convenience functions."""

from gettext import gettext as _

import datetime
import math

import lazy_import

from pedantic_timedelta import PedanticTimedelta

# Profiling: load pytz: ~ 0.002 secs.
pytz = lazy_import.lazy_module('pytz')


__all__ = (
    'isoformat',
    'isoformat_tzinfo',
    'isoformat_tzless',
)


# ***

def isoformat(dt, sep='T', timespec='auto', include_tz=False):
    """
    FIXME: Document

    Based loosely on::

        datetime.isoformat(sep='T', timespec='auto')

    in Python 3.6 (which added timespec).

    The optional argument sep (default 'T') is a one-character separator,
    placed between the date and time portions of the result.

    The optional argument timespec specifies the number of additional components
    of the time to include (the default is 'auto'). It can be one of the following:

    'auto': Same as 'seconds' if microsecond is 0, same as 'microseconds' otherwise.
    'hours': Include the hour in the two-digit HH format.
    'minutes': Include hour and minute in HH:MM format.
    'seconds': Include hour, minute, and second in HH:MM:SS format.
    'milliseconds': Include full time, but truncate fractional second part
    to milliseconds. HH:MM:SS.sss format.
    'microseconds': Include full time in HH:MM:SS.mmmmmm format.

    Note: Excluded time components are truncated, not rounded.

    ValueError will be raised on an invalid timespec argument.

    """
    def _isoformat(dt, sep, timespec, include_tz):
        timecomp = _format_timespec(dt, timespec)

        tzcomp = ''
        if dt.tzinfo:
            if include_tz:
                tzcomp = '%z'
            else:
                dt = dt.astimezone(pytz.utc)
        # else, a naive datetime, we'll just have to assume it's UTC!

        return dt.strftime('%Y-%m-%d{}{}{}'.format(sep, timecomp, tzcomp))

    def _format_timespec(dt, timespec):
        if timespec == 'auto':
            if not dt.microsecond:
                timespec = 'seconds'
            else:
                timespec = 'microseconds'

        if timespec == 'hours':
            return '%H'
        elif timespec == 'minutes':
            return '%H:%M'
        elif timespec == 'seconds':
            return '%H:%M:%S'
        elif timespec == 'milliseconds':
            msec = '{:03}'.format(math.floor(dt.microsecond / 1000))
            return '%H:%M:%S.{}'.format(msec)
        elif timespec == 'microseconds':
            return '%H:%M:%S.%f'
        else:
            raise ValueError('Not a valid `timespec`: {}'.format(timespec))

    return _isoformat(dt, sep, timespec, include_tz)


def isoformat_tzinfo(dt, sep='T', timespec='auto'):
    """FIXME: Document"""
    if isinstance(dt, datetime.datetime):
        return isoformat(dt, sep=sep, timespec=timespec, include_tz=True)
    else:
        return dt


def isoformat_tzless(dt, sep='T', timespec='auto'):
    """FIXME: Document"""
    if isinstance(dt, datetime.datetime):
        return isoformat(dt, sep=sep, timespec=timespec, include_tz=False)
    else:
        return dt


# ***

def format_delta(delta, style='%M'):
    """
    Return a string representation of ``Fact().delta``.

    Args:
        style (str): Specifies the output format.

          Valid choices are:
            * ``'%M'``: As minutes, rounded down.
            * ``'%H:%M'``: As 'hours:minutes'. rounded down.
            * ``HHhMMm``: As '{hours} hour(s) {minutes} minute(s)'.
            * ````: As human friendly time.

    Returns:
        str: Formatted string representing this fact's *duration*.
    """
    def _format_delta():
        seconds = delta.total_seconds() if delta is not None else 0
        if not style:
            return format_pedantic(seconds)
        elif style == '%M':
            minutes = int(seconds / 60)
            return format_mins(minutes)
        else:
            hours = int(seconds / 3600)
            minutes = int((seconds % 3600) / 60)
            if style == '%H:%M':
                return format_hours_mins(hours, minutes)
            elif style == 'HHhMMm':
                return format_hours_h_mins_m(hours, minutes)
        raise ValueError(_("Invalid format_delta style ‘{}’.").format(style))

    def format_mins(minutes):
        return str(minutes)

    def format_hours_mins(hours, minutes):
        return '{0:02d}:{1:02d}'.format(hours, minutes)

    def format_hours_h_mins_m(hours, minutes):
        text = ''
        text += "{0:>2d} ".format(hours)
        text += _("hour ") if hours == 1 else _("hours")
        text += " {0:>2d} ".format(minutes)
        text += _("minute ") if minutes == 1 else _("minutes")
        return text

    def format_pedantic(seconds):
        (
            tm_fmttd, tm_scale, tm_units,
        ) = PedanticTimedelta(seconds=seconds).time_format_scaled()
        return tm_fmttd

    return _format_delta()




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.


class TestAlchemyCategory(object):
    """Make sure our custom methods behave properly."""
    def test_as_hamster(self, alchemy_store, alchemy_category):
        """Make sure that conversion into a ``nark.Category```works as expected."""
        category = alchemy_category.as_hamster(alchemy_store)
        assert category.equal_fields(alchemy_category)


class TestAlchemyActivity(object):
    """Make sure our custom methods behave properly."""
    def test_as_hamster(self, alchemy_store, alchemy_activity):
        """Make sure that conversion into a ``nark.Activity```works as expected."""
        activity = alchemy_activity.as_hamster(alchemy_store)
        assert activity.equal_fields(alchemy_activity)

    def test_activity_has_facts(self, alchemy_store, alchemy_fact_factory):
        """Make sure that an activity can access ``Fact`` instances."""
        alchemy_fact = alchemy_fact_factory()
        assert alchemy_fact.activity
        activity = alchemy_fact.activity
        assert activity.facts


class TestAlchemyTag(object):
    """Make sure our custom methods behave properly."""
    def test_as_hamster(self, alchemy_store, alchemy_tag):
        """Make sure that conversion into a ``nark.Tag``works as expected."""
        tag = alchemy_tag.as_hamster(alchemy_store)
        assert tag.equal_fields(alchemy_tag)


class TestAlchemyFact(object):
    """Make sure our custom methods behave properly."""

    def test_adding_tags(self, alchemy_store, alchemy_fact, alchemy_tag):
        """
        Make sure that adding tags works as expected.

        This is closer to testing we got SQLAlchemy right than actual code.
        """
        assert len(alchemy_fact.tags) == 4
        alchemy_fact.tags.append(alchemy_tag)
        assert len(alchemy_fact.tags) == 5
        assert alchemy_tag in alchemy_fact.tags

    def test_setting_tags(self, alchemy_store, alchemy_fact, alchemy_tag_factory):
        """
        Make sure that adding tags works as expected.

        This is closer to testing we got SQLAlchemy right than actual code.
        """
        assert alchemy_fact.tags
        new_tags = [alchemy_tag_factory() for i in range(5)]
        alchemy_fact.tags = new_tags
        assert len(alchemy_fact.tags) == 5
        assert alchemy_fact.tags == new_tags

    def test_as_hamster(self, alchemy_store, alchemy_fact):
        """Make sure that conversion into a ``nark.Fact```works as expected."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        assert fact.equal_fields(alchemy_fact)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Provides string-related functional methods."""

__all__ = (
    'format_value_truncate',
)


def format_value_truncate(val, trunc_width=None):
    if not val:
        return val
    # (lb): First attempt was to only short first line [of Fact description]:
    #   vals = val.splitlines()
    #   if len(vals) > 1:
    #       val = vals[0] + '...'
    # but replacing newlines with representation of same
    # is more meaningful, and less confusing to end user.
    val = '\\n'.join(str(val).splitlines())
    if trunc_width is not None:
        if len(val) > trunc_width and trunc_width >= 0:
            val = val[:trunc_width - 3] + '...'
    return val




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import logging

import pytest

from nark import get_version
from nark.manager import BaseStore


class TestController:
    @pytest.mark.parametrize('storetype', ['sqlalchemy'])
    def test_get_store_valid(self, controller, storetype):
        """Make sure we receive a valid ``store`` instance."""
        # [TODO]
        # Once we got backend registration up and running this should be
        # improved to check actual store type for that backend.
        controller.config['db']['orm'] = storetype
        assert isinstance(controller._get_store(), BaseStore)

    def test_get_store_invalid(self, controller):
        """Make sure we get an exception if store retrieval fails."""
        # Because of config validation, as opposed to hamster-lib,
        # you cannot set the config to bad values, i.e., the code
        # fails before we're able to call, says, controller._get_store().
        with pytest.raises(ValueError):
            controller.config['db']['orm'] = None

    def test_update_config(self, controller, base_config, mocker):
        """Make sure we assign new config and get a new store."""
        controller._get_store = mocker.MagicMock()
        controller.update_config(base_config)
        assert controller.config.as_dict() == base_config
        assert controller._get_store.called

    def test_get_logger(self, controller):
        """Make sure we recieve a logger that maches our expectations."""
        logger = controller._get_logger()
        assert isinstance(logger, logging.Logger)
        assert logger.name == 'nark.log'
        # [FIXME]
        # assert len(logger.handlers) == 1
        assert isinstance(logger.handlers[0], logging.NullHandler)

    def test_sql_logger(self, controller):
        """Make sure we recieve a logger that maches our expectations."""
        logger = controller._sql_logger()
        assert isinstance(logger, logging.Logger)
        assert logger.name == 'nark.store'
        assert isinstance(logger.handlers[0], logging.NullHandler)


class TestNarkLib:
    def test_get_version(self):
        # (lb): Not sure how best to test get_version, because it
        # behaves differently if setuptools_scm is included or not,
        # and the version will often be a non-release version, e.g.,
        # '3.0.2.dev9+gfba2058.d20200401'. For now, just say not empty.
        assert get_version() != ''




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""This module provides nark raw fact parsing-related functions."""

__all__ = (
    'ParserException',
    'ParserMissingDatetimeException',
    'ParserMissingDatetimeOneException',
    'ParserMissingDatetimeTwoException',
    'ParserInvalidDatetimeException',
    'ParserMissingSeparatorActivity',
    'ParserMissingActivityException',
)


class ParserException(Exception):
    """Raised if parser cannot decipher nark factoid string."""
    pass


class ParserMissingDatetimeException(ParserException):  # noqa: E302
    """Raised if the raw_fact is missing one or both datetime tokens."""
    pass


class ParserMissingDatetimeOneException(ParserMissingDatetimeException):  # noqa: E302
    """Raised if the raw_fact is missing its start datetime token(s)."""
    pass


class ParserMissingDatetimeTwoException(ParserMissingDatetimeException):  # noqa: E302
    """Raised if the raw_fact is missing its end datetime token(s)."""
    pass


class ParserInvalidDatetimeException(ParserException):  # noqa: E302
    """Raised if a time from raw_fact in not parseworthy."""
    pass


class ParserMissingSeparatorActivity(ParserException):  # noqa: E302
    """Raised if activity@category separator not found."""
    pass


class ParserMissingActivityException(ParserException):  # noqa: E302
    """Raised if factoid is missing: act@cat, cat@, @cat, or just @."""
    pass




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from collections import namedtuple

from .category import Category
from .item_base import BaseItem

ActivityTuple = namedtuple(
    'ActivityTuple', ('pk', 'name', 'category', 'deleted', 'hidden'),
)


class Activity(BaseItem):
    """Storage agnostic class for activities."""

    def __init__(self, name, pk=None, category=None, deleted=False, hidden=False):
        """
        Initialize this instance.

        Args:
            name (str): This is the name of the activity. May contain whitespace!

            pk: The unique primary key used by the backend.

            category (Category): ``Category`` instance associated with this ``Activity``.

            deleted (bool): True if this ``Activity`` has been marked as deleted.

        Note:
            *Legacy hamster* basically treated ``(Activity.name, Category.name)`` as
            *composite keys*. As a consequence ``Activity.names`` themselves are not
            unique. They are only in combination with their associated categories name.
        """
        # [TODO]
        # Elaborate on the consequences of the deleted flag.

        super(Activity, self).__init__(pk, name)
        self.category = category
        self.deleted = bool(deleted)
        self.hidden = bool(hidden)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        # NOTE: (lb): Unlike Category.name=, not requiring ``not name``.
        if name is None:
            raise ValueError(_('Activity name must not be None.'))
        self._name = str(name)

    @classmethod
    def create_from_composite(cls, name, category_name, deleted=False):
        """
        Convenience method that allows creating a new instance providing
            the 'natural key'.

        Args:
            name (str): This activities name.
            category_name (str): Name of the associated category.
            deleted (bool): True if this ``Activity`` has been marked as deleted.

        Returns:
            Activity: A new ``Activity`` instance.

        Note:
            * Should future iterations extend ``Category`` this may turn problematic.
            * This method does not allow to specify a primary key as it is intended
              only for new instances, not ones retrieved by the backend.

        """
        category = Category(category_name)
        return cls(name, category=category, deleted=deleted)

    def as_tuple(self, include_pk=True):
        """
        Provide a tuple representation of this activities relevant 'fields'.

        Args:
            include_pk (bool): Whether to include the instances pk or not.
                Note that if ``False`` ``tuple.pk = False``!

        Returns:
            ActivityTuple: Representing this activities values.
        """
        pk = self.pk
        if not include_pk:
            pk = False
        if self.category:
            category = self.category.as_tuple(include_pk=include_pk)
        else:
            category = None
        return ActivityTuple(
            pk=pk,
            name=self.name,
            category=category,
            deleted=bool(self.deleted),
            hidden=bool(self.hidden),
        )

    def equal_fields(self, other):
        """
        Compare this instances fields with another activity.
            This excludes comparing the PK.

        Args:
            other (Activity): Activity to compare this instance with.

        Returns:
            bool: ``True`` if all fields but ``pk`` are equal, ``False`` if not.

        Note:
            This is particularly useful if you want to compare a new ``Activity``
            instance with a freshly created backend instance. As the latter will
            probably have a primary key assigned now and so ``__eq__`` would fail.
        """
        return self.as_tuple(include_pk=False) == other.as_tuple(include_pk=False)

    def __eq__(self, other):
        if other is not None and not isinstance(other, ActivityTuple):
            other = other.as_tuple()
        return self.as_tuple() == other

    def __hash__(self):
        """Naive hashing method."""
        return hash(self.as_tuple())

    def __str__(self):
        if self.category is None:
            full_name = '{name}'.format(name=self.name)
        else:
            full_name = '{name} ({category})'.format(
                name=self.name, category=self.category.name,
            )
        return full_name




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from sqlalchemy import asc, desc, func
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound

from . import BaseAlchemyManager, query_apply_limit_offset, query_apply_true_or_not
from ....managers.activity import BaseActivityManager
from ..objects import AlchemyActivity, AlchemyCategory, AlchemyFact


class ActivityManager(BaseAlchemyManager, BaseActivityManager):
    def get_or_create(self, activity, raw=False, skip_commit=False):
        """
        Custom version of the default method in order to provide access to
        Alchemy instances.

        Args:
            activity (nark.Activity): Activity we want.
            raw (bool): Wether to return the AlchemyActivity instead.

        Returns:
            nark.Activity: Activity.
        """

        message = _("Received {!r}, raw={}.".format(activity, raw))
        self.store.logger.debug(message)

        try:
            result = self.get_by_composite(activity.name, activity.category, raw=raw)
        except KeyError:
            result = self._add(activity, raw=raw, skip_commit=skip_commit)
        self.store.logger.debug(_("Returning {!r}.").format(result))
        return result

    def _add(self, activity, raw=False, skip_commit=False):
        """
        Add a new ``Activity`` instance to the databasse.

        Args:
            activity (nark.Activity): nark activity

        Returns:
            nark.Activity: nark activity representation of stored instance.

        Raises:
            ValueError: If the passed activity has a PK.
            ValueError: If the category/activity.name combination to be added is
                already present in the db.
        """
        self.adding_item_must_not_have_pk(activity)

        try:
            self.get_by_composite(activity.name, activity.category)
            # FIXME/2018-06-08: (lb): DRY: See "Our database already" elsewhere.
            message = _("Our database already contains the passed name/category.name"
                        "combination.")
            self.store.logger.error(message)
            raise ValueError(message)
        except KeyError:
            pass

        alchemy_activity = AlchemyActivity(
            pk=None,
            name=activity.name,
            category=None,
            deleted=bool(activity.deleted),
            hidden=bool(activity.hidden),
        )
        if activity.category:
            try:
                category = self.store.categories.get_by_name(
                    activity.category.name, raw=True)
            except KeyError:
                category = AlchemyCategory(
                    pk=None,
                    name=activity.category.name,
                    deleted=bool(activity.category.deleted),
                    hidden=bool(activity.category.hidden),
                )
        else:
            category = None
        alchemy_activity.category = category

        result = self.add_and_commit(
            alchemy_activity, raw=raw, skip_commit=skip_commit,
        )

        return result

    def _update(self, activity):
        """
        Update a given Activity.

        Args:
            activity (nark.Activity): Activity to be updated.

        Returns:
            nark.Activity: Updated activity.

        Raises:
            ValueError: If the new name/category.name combination is already taken.
            ValueError: If the the passed activity does not have a PK assigned.
            KeyError: If the the passed activity.pk can not be found.
        """

        message = _("Received {!r}.".format(activity))
        self.store.logger.debug(message)

        if not activity.pk:
            message = _(
                "The activity passed ('{!r}') does not seem to havea PK. We don't know"
                "which entry to modify.".format(activity))
            self.store.logger.error(message)
            raise ValueError(message)

        try:
            self.get_by_composite(activity.name, activity.category)
            # FIXME/2018-06-08: (lb): DRY: See "Our database already" elsewhere.
            message = _("Our database already contains the passed name/category.name"
                        "combination.")
            self.store.logger.error(message)
            raise ValueError(message)
        except KeyError:
            pass

        alchemy_activity = self.store.session.query(AlchemyActivity).get(activity.pk)
        if not alchemy_activity:
            message = _("No activity with this pk can be found.")
            self.store.logger.error(message)
            raise KeyError(message)
        alchemy_activity.name = activity.name
        alchemy_activity.category = self.store.categories.get_or_create(
            activity.category, raw=True,
        )
        alchemy_activity.deleted = bool(activity.deleted)
        try:
            self.store.session.commit()
        except IntegrityError as e:
            message = _(
                'There seems to already be an activity like this for the given category.'
                " Cannot change this activity's values. Original exception: {}".format(e)
            )
            self.store.logger.error(message)
            raise ValueError(message)
        result = alchemy_activity.as_hamster(self.store)
        self.store.logger.debug(_("Returning: {!r}.".format(result)))
        return result

    def remove(self, activity):
        """
        Remove an activity from our internal backend.

        Args:
            activity (nark.Activity): The activity to be removed.

        Returns:
            bool: True

        Raises:
            KeyError: If the given ``Activity`` can not be found in the database.
        """

        message = _("Received {!r}.".format(activity))
        self.store.logger.debug(message)

        if not activity.pk:
            message = _(
                "The activity you passed does not have a PK. Please provide one."
            )
            self.store.logger.error(message)
            raise ValueError(message)

        alchemy_activity = self.store.session.query(AlchemyActivity).get(activity.pk)
        if not alchemy_activity:
            message = _("The activity you try to remove does not seem to exist.")
            self.store.logger.error(message)
            raise KeyError(message)
        if alchemy_activity.facts:
            alchemy_activity.deleted = True
            self.store.activities._update(alchemy_activity)
        else:
            self.store.session.delete(alchemy_activity)
        self.store.session.commit()
        self.store.logger.debug(_("Deleted {!r}.".format(activity)))
        return True

    def get(self, pk, deleted=None, raw=False):
        """
        Query for an Activity with given key.

        Args:
            pk: PK to look up.
            raw (bool): Return the AlchemyActivity instead.

        Returns:
            nark.Activity: Activity with given PK.

        Raises:
            KeyError: If no such pk was found.
        """

        message = _("Received PK: '{}', raw={}.".format(pk, raw))
        self.store.logger.debug(message)

        if deleted is None:
            result = self.store.session.query(AlchemyActivity).get(pk)
        else:
            query = self.store.session.query(AlchemyActivity)
            query = query.filter(AlchemyActivity.pk == pk)
            query = query_apply_true_or_not(query, AlchemyActivity.deleted, deleted)
            results = query.all()
            assert(len(results) <= 1)
            result = results[0] if results else None

        if not result:
            message = _("No Activity with 'pk: {}' was found!".format(pk))
            self.store.logger.error(message)
            raise KeyError(message)
        if not raw:
            result = result.as_hamster(self.store)
        self.store.logger.debug(_("Returning: {!r}.".format(result)))
        return result

    def get_by_composite(self, name, category, raw=False):
        """
        Retrieve an activity by its name and category.

        Args:
            name (str): The activities name.
            category (nark.Category or None): The activities category.
                May be None.
            raw (bool): Return the AlchemyActivity instead.

        Returns:
            nark.Activity: The activity if it exists in this combination.

        Raises:
            KeyError: if composite key can not be found in the db.

        Note:
            As far as we understand the legacy code in ``__change_category``
            and ``__get_activity_by`` the combination of activity.name and
            activity.category is unique. This is reflected in the uniqueness
            constraint of the underlying table.
        """

        message = _(
            "Received name: '{}' and {!r} with 'raw'={}.".format(name, category, raw)
        )
        self.store.logger.debug(message)

        if category:
            category = category.name
            try:
                alchemy_category = self.store.categories.get_by_name(category, raw=True)
            except KeyError:
                message = _(
                    'The category passed ({}) does not exist in the backend. '
                    'Consequently no related activity can be returned.'
                    .format(category)
                )
                # (lb): This was error, but shouldn't be; callers catch if they care.
                self.store.logger.debug(message)
                raise KeyError(message)
        else:
            alchemy_category = None

        # EXPLAIN: (lb): Is name ever not a string here?
        name = str(name)
        try:
            query = self.store.session.query(AlchemyActivity)
            query = query.filter_by(name=name).filter_by(category=alchemy_category)
            result = query.one()
        except NoResultFound:
            message = _(
                "No activity named '{name}' of category '{category}' was found"
                .format(name=name, category=category)
            )
            self.store.logger.debug(message)
            raise KeyError(message)
        if not raw:
            result = result.as_hamster(self.store)
        self.store.logger.debug(_("Returning: {!r}.".format(result)))
        return result

    def get_all(self, *args, include_usage=False, sort_col='name', **kwargs):
        """Get all activities."""
        kwargs['include_usage'] = include_usage
        kwargs['sort_col'] = sort_col
        return self._get_all(*args, **kwargs)

    def get_all_by_usage(self, *args, sort_col='usage', **kwargs):
        assert(not args)
        kwargs['include_usage'] = True
        kwargs['sort_col'] = sort_col
        return self._get_all(*args, **kwargs)

    def _get_all(
        self,
        include_usage=True,
        count_results=False,
        # FIXME/2018-06-20: (lb): Implement since/until.
        since=None,
        until=None,
        # FIXME/2018-06-09: (lb): Implement deleted/hidden.
        deleted=False,
        hidden=False,
        # FIXME/2018-06-20: (lb): Do what with key now?
        key=None,
        search_term='',
        category=False,
        activity=False,
        sort_col='',
        sort_order='',
        # kwargs: limit, offset
        **kwargs
    ):
        """
        FIXME: Update this docstring.

        Retrieve all matching activities stored in the backend.

        Args:
            include_usage (int, optional): If true, include count of Facts that reference
                each Activity.
            search_term (str, optional): Limit activities to those matching a substring
                in their name. Defaults to ``empty string``.
            category (nark.Category or str, optional): Limit activities to this
                category. Defaults to ``False``. If ``category=None`` only activities
                without a category will be considered.
            activity (nark.Activity, optional): Limit activities to this activity.
                Defaults to ``False``. If ``activity=None`` only activities with a
                matching name will be considered.
            sort_col (str, optional): Which columns to sort by. Defaults to 'activity'.
                Choices: 'activity, 'category', 'start', 'usage'.
                Note that 'start' and 'usage' only apply if include_usage.
            sort_order (str, optional): One of:
                'asc': Whether to search the results in ascending order.
                'desc': Whether to search the results in descending order.
            limit (int, optional): Query "limit".
            offset (int, optional): Query "offset".

        Returns:
            list: List of ``nark.Activity`` instances matching constrains.
                The list is ordered by ``Activity.name``.
        """

        def _get_all_activities():
            message = (
                _('usage: {} / term: {} / cat.: {} / act.: {} / col: {} / order: {}')
                .format(
                    include_usage, search_term, category, activity, sort_col, sort_order,
                )
            )
            self.store.logger.debug(message)

            query, agg_cols = _get_all_start_query()

            query = _get_all_filter_by_category(query)

            query = _get_all_filter_by_activity(query)

            query = _get_all_filter_by_search_term(query)

            query = _get_all_order_by(query, *agg_cols)

            query = _get_all_group_by(query, agg_cols)

            query = query_apply_limit_offset(query, **kwargs)

            query = _get_all_with_entities(query, agg_cols)

            self.store.logger.debug(_('Query') + ': {}'.format(str(query)))

            results = query.all() if not count_results else query.count()

            return results

        def _get_all_start_query():
            agg_cols = []
            if not include_usage:
                query = self.store.session.query(AlchemyActivity)
            else:
                count_col = func.count(AlchemyActivity.pk).label('uses')
                agg_cols.append(count_col)

                time_col = func.sum(
                    func.julianday(AlchemyFact.end) - func.julianday(AlchemyFact.start)
                ).label('span')
                agg_cols.append(time_col)

                query = self.store.session.query(AlchemyFact, count_col, time_col)
                query = query.join(AlchemyFact.activity)
                # NOTE: (lb): SQLAlchemy will lazy load category if/when caller
                #       references it. (I tried using query.options(joinedload(...))
                #       but that route was a mess; and I don't know SQLAlchemy well.)

            # SQLAlchemy automatically lazy-loads activity.category if we
            # reference it after executing the query, so we don't need to
            # join, except that we want to sort by category.name, so we do.
            query = query.join(AlchemyCategory)

            return query, agg_cols

        def _get_all_filter_by_category(query):
            if category is False:
                return query
            # FIXME/2018-05-15 22:30: (lb): MRU activities by category: GLOB:
            #   do fuzzy search on name rather than PK, if special query,
            #     e.g., ``if '*' in category.name: ...``
            if category:
                category_name = None
                if isinstance(category, str):
                    category_name = category
                elif not category.pk:
                    category_name = category.name

                if not category_name:
                    query = query.filter(AlchemyCategory.pk == category.pk)
                else:
                    query = query.filter(
                        func.lower(AlchemyCategory.name) == func.lower(category_name)
                    )
            else:
                query = query.filter(AlchemyActivity.category == None)  # noqa: E711
            return query

        def _get_all_filter_by_activity(query):
            if activity is False:
                return query
            if activity:
                if activity.pk:
                    query = query.filter(AlchemyActivity.pk == activity.pk)
                else:
                    query = query.filter(
                        func.lower(AlchemyActivity.name) == func.lower(activity.name)
                    )
            else:
                query = query.filter(AlchemyFact.activity == None)  # noqa: E711
            return query

        def _get_all_filter_by_search_term(query):
            if not search_term:
                return query
            query = query.filter(
                AlchemyActivity.name.ilike('%{}%'.format(search_term))
            )
            return query

        def _get_all_order_by(query, count_col=None, time_col=None):
            direction = desc if sort_order == 'desc' else asc
            if sort_col == 'start':
                assert include_usage
                direction = desc if not sort_order else direction
                query = query.order_by(direction(AlchemyFact.start))
            elif sort_col == 'usage':
                assert include_usage and count_col is not None
                direction = desc if not sort_order else direction
                query = query.order_by(direction(count_col), direction(time_col))
            elif sort_col == 'time':
                assert include_usage and time_col is not None
                direction = desc if not sort_order else direction
                query = query.order_by(direction(time_col), direction(count_col))
            elif sort_col == 'activity':
                query = query.order_by(direction(AlchemyActivity.name))
                query = query.order_by(direction(AlchemyCategory.name))
            elif sort_col == 'category':
                query = query.order_by(direction(AlchemyCategory.name))
                query = query.order_by(direction(AlchemyActivity.name))
            else:
                assert sort_col in ('', 'name', 'tag', 'fact')
                query = query.order_by(direction(AlchemyActivity.name))
                query = query.order_by(direction(AlchemyCategory.name))
            return query

        def _get_all_group_by(query, agg_cols):
            if not agg_cols:
                return query
            query = query.group_by(AlchemyActivity.pk)
            return query

        def _get_all_with_entities(query, agg_cols):
            if not agg_cols:
                return query
            # (lb): Get tricky with it. The query now SELECTs all Fact columns,
            #  and it JOINs and GROUPs BY activities to produce the counts. But
            #  the Fact is meaningless after the group-by; we want the Activity
            #  instead. So use with_entities trickery to tell SQLAlchemy which
            #  columns we really want -- it'll transform the query so that the
            #  SELECT fetches all the Activity columns; but the JOIN and GROUP BY
            #  remain the same! (I'm not quite sure how it works, but it does.)
            # And as an aside, because of the 1-to-many relationship, the
            #  Activity table does not reference Fact, so, e.g., this wouldn't
            #  work, or at least I assume not, but maybe SQLAlchemy would figure
            #  it out: self.store.session.query(AlchemyActivity).join(AlchemyFact).
            query = query.with_entities(AlchemyActivity, *agg_cols)
            return query

        # ***

        return _get_all_activities()




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from datetime import datetime

from sqlalchemy import asc, desc, func
from sqlalchemy.sql.expression import and_, or_

from . import BaseAlchemyManager, query_apply_limit_offset, query_apply_true_or_not
from ....managers.fact import BaseFactManager
from ..objects import (
    AlchemyActivity,
    AlchemyCategory,
    AlchemyFact,
    AlchemyTag,
    fact_tags
)


class FactManager(BaseAlchemyManager, BaseFactManager):
    """
    """
    def __init__(self, *args, **kwargs):
        super(FactManager, self).__init__(*args, **kwargs)

    # ***

    def _timeframe_available_for_fact(self, fact, ignore_pks=[]):
        """
        Determine if a timeframe given by the passed fact is already occupied.

        This method takes also such facts into account that start before and end
        after the fact in question. In that regard it exceeds what ``_get_all``
        would return.

        Args:
            fact (Fact): The fact to check. Please note that the fact is expected to
                have a ``start`` and ``end``.

        Returns:
            bool: ``True`` if the timeframe is available, ``False`` if not.

        Note:
            If the given fact is the only fact instance within the given timeframe
            the timeframe is considered available (for this fact)!
        """
        # Use func.datetime and _get_sql_datetime to normalize time comparisons,
        # so that equivalent times that are expressed differently are evaluated
        # as equal, e.g., "2018-01-01 10:00" should match "2018-01-01 10:00:00".
        # FIXME: func.datetime is SQLite-specific: need to abstract for other DBMSes.

        start = self._get_sql_datetime(fact.start)
        query = self.store.session.query(AlchemyFact)

        # FIXME: Only use func.datetime on SQLite store.
        #
        #   (lb): SQLite stores datetimes as strings, so what's in the store
        #   might vary depending on, say, changes to this code. As such, some
        #   start and end times might include seconds, and some times might not.
        #   Here we use func.datetime and _get_sql_datetime to normalize the
        #   comparison. But this is SQLite-specific, so we should abstract
        #   the operation for other DBMSes (and probably do nothing, since most
        #   other databases have an actual datetime data type).
        condition = and_(func.datetime(AlchemyFact.end) > start)
        if fact.end is not None:
            end = self._get_sql_datetime(fact.end)
            condition = and_(condition, func.datetime(AlchemyFact.start) < end)
        else:
            # The fact is ongoing, so also match the ongoing Fact in the store.
            # E711: `is None` breaks Alchemy, so use `== None`.
            condition = or_(AlchemyFact.end == None, condition)  # noqa: E711

        if fact.pk:
            condition = and_(condition, AlchemyFact.pk != fact.pk)

        if fact.split_from:
            condition = and_(condition, AlchemyFact.pk != fact.split_from.pk)

        if ignore_pks:
            condition = and_(condition, AlchemyFact.pk.notin_(ignore_pks))

        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712

        query = query.filter(condition)

        return not bool(query.count())

    # ***

    def _add(self, fact, raw=False, skip_commit=False, ignore_pks=[]):
        """
        Add a new fact to the database.

        Args:
            fact (nark.Fact): Fact to be added.
            raw (bool): If ``True`` return ``AlchemyFact`` instead.

        Returns:
            nark.Fact: Fact as stored in the database

        Raises:
            ValueError: If the passed fact has a PK assigned.
                New facts should not have one.

            ValueError: If the timewindow is already occupied.
        """
        self.adding_item_must_not_have_pk(fact)

        self.must_validate_datetimes(fact, ignore_pks=ignore_pks)

        alchemy_fact = AlchemyFact(
            pk=None,
            activity=None,
            # FIXME/2018-08-23 00:38: Is this still valid?
            # FIXME: mircoseconds are being stored...
            #        I modified fact_time.must_be_datetime_or_relative to strip
            #        milliseconds. but they're still being saved (just as six 0's).
            start=fact.start,
            end=fact.end,
            description=fact.description,
            deleted=bool(fact.deleted),
            split_from=fact.split_from,
        )
        get_or_create = self.store.activities.get_or_create
        alchemy_fact.activity = get_or_create(fact.activity, raw=True, skip_commit=True)
        tags = [
            self.store.tags.get_or_create(tag, raw=True, skip_commit=True)
            for tag in fact.tags
        ]
        alchemy_fact.tags = tags

        result = self.add_and_commit(
            alchemy_fact, raw=raw, skip_commit=skip_commit,
        )

        return result

    # ***

    def _update(self, fact, raw=False, ignore_pks=[]):
        """
        Update and existing fact with new values.

        Args:
            fact (nark.fact): Fact instance holding updated values.

            raw (bool): If ``True`` return ``AlchemyFact`` instead.
              ANSWER: (lb): "instead" of what? raw is not used by Fact...

        Returns:
            nark.fact: Updated Fact

        Raises:
            KeyError: if a Fact with the relevant PK could not be found.
            ValueError: If the the passed activity does not have a PK assigned.
            ValueError: If the timewindow is already occupied.
        """
        self.store.logger.debug(_("Received '{!r}', 'raw'={}.".format(fact, raw)))

        if not fact.pk:
            message = _(
                "{!r} does not seem to have a PK. We don't know"
                "which entry to modify.".format(fact)
            )
            self.store.logger.error(message)
            raise ValueError(message)

        self.must_validate_datetimes(fact, ignore_pks=ignore_pks)

        alchemy_fact = self.store.session.query(AlchemyFact).get(fact.pk)
        if not alchemy_fact:
            message = _("No fact with PK: {} was found.".format(fact.pk))
            self.store.logger.error(message)
            raise KeyError(message)

        if alchemy_fact.deleted:
            message = _('Cannot edit deleted Fact: {!r}'.format(fact))
            self.store.logger.error(message)
            raise ValueError(message)

        if (
            (
                (fact.deleted and (fact.end == alchemy_fact.end))
                or (fact.end and not alchemy_fact.end)
            )
            and fact.equal_sans_end(alchemy_fact)
        ):
            # Don't bother with split_from entry.
            # MAYBE: (lb): Go full wiki and store edit times? Ug...
            new_fact = alchemy_fact
            alchemy_fact.deleted = fact.deleted
            alchemy_fact.end = fact.end
        else:
            assert alchemy_fact.pk == fact.pk
            was_split_from = fact.split_from
            fact.split_from = alchemy_fact
            # Clear the ID so that a new ID is assigned.
            fact.pk = None
            new_fact = self._add(fact, raw=True, skip_commit=True, ignore_pks=ignore_pks)
            # NOTE: _add() calls:
            #       self.store.session.commit()
            # The fact being split from is deleted/historic.
            alchemy_fact.deleted = True
            assert new_fact.pk > alchemy_fact.pk
            # Restore the ID to not confuse the caller!
            # The caller will still have a handle on Fact. Rather than
            # change its pk to new_fact's, have it reflect its new
            # split_from status.
            fact.pk = alchemy_fact.pk
            fact.split_from = was_split_from
            # The `alchemy_fact` is what gets saved, but the `fact`
            # is what the caller passed us, so update it, too.
            fact.deleted = True

        self.store.session.commit()

        self.store.logger.debug(_("Updated {!r}.".format(fact)))

        if not raw:
            new_fact = new_fact.as_hamster(self.store)

        return new_fact

    # ***

    def must_validate_datetimes(self, fact, ignore_pks=[]):
        if not isinstance(fact.start, datetime):
            raise TypeError(_('Missing start time for “{!r}”.').format(fact))

        # Check for valid time range.
        invalid_range = False
        if fact.end is not None:
            if fact.start > fact.end:
                invalid_range = True
            else:
                # EXPERIMENTAL: Sneaky, "hidden", vacant, timeless Facts.
                allow_momentaneous = self.store.config['time.allow_momentaneous']
                if not allow_momentaneous and fact.start >= fact.end:
                    invalid_range = True

        if invalid_range:
            message = _('Invalid time range for “{!r}”.').format(fact)
            if fact.start == fact.end:
                assert False  # (lb): Preserved in case we revert == policy.
                message += _(' The start time cannot be the same as the end time.')
            else:
                message += _(' The start time cannot be after the end time.')
            self.store.logger.error(message)
            raise ValueError(message)

        if not self._timeframe_available_for_fact(fact, ignore_pks):
            msg = _(
                'One or more Facts already exist '
                'between the indicated start and end times. '
            )
            self.store.logger.error(msg)
            raise ValueError(msg)

    # ***

    def remove(self, fact, purge=False):
        """
        Remove a fact from our internal backend.

        Args:
            fact (nark.Fact): Fact to be removed

        Returns:
            bool: Success status

        Raises:
            ValueError: If fact passed does not have an pk.

            KeyError: If no fact with passed PK was found.
        """
        self.store.logger.debug(_("Received '{!r}'.".format(fact)))

        if not fact.pk:
            message = _(
                "The fact passed ('{!r}') does not seem to havea PK. We don't know"
                "which entry to remove.".format(fact)
            )
            self.store.logger.error(message)
            raise ValueError(message)

        alchemy_fact = self.store.session.query(AlchemyFact).get(fact.pk)
        if not alchemy_fact:
            message = _('No fact with given pk was found!')
            self.store.logger.error(message)
            raise KeyError(message)
        if alchemy_fact.deleted:
            message = _('The Fact is already marked deleted.')
            self.store.logger.error(message)
            # FIXME/2018-06-08: (lb): I think we need custom Exceptions...
            raise Exception(message)
        alchemy_fact.deleted = True
        if purge:
            self.store.session.delete(alchemy_fact)
        self.store.session.commit()
        self.store.logger.debug(_('{!r} has been removed.'.format(fact)))
        return True

    # ***

    def get(self, pk, deleted=None, raw=False):
        """
        Retrieve a fact based on its PK.

        Args:
            pk (int): PK of the fact to be retrieved.

            deleted (boolean, optional):
                False to restrict to non-deleted Facts;
                True to find only those marked deleted;
                None to find all.

            raw (bool): Return the AlchemyActivity instead.

        Returns:
            nark.Fact: Fact matching given PK

        Raises:
            KeyError: If no Fact of given key was found.
        """
        self.store.logger.debug(_("Received PK: {}', 'raw'={}.".format(pk, raw)))

        if deleted is None:
            result = self.store.session.query(AlchemyFact).get(pk)
        else:
            query = self.store.session.query(AlchemyFact)
            query = query.filter(AlchemyFact.pk == pk)
            query = query_apply_true_or_not(query, AlchemyFact.deleted, deleted)
            results = query.all()
            assert(len(results) <= 1)
            result = results[0] if results else None

        if not result:
            message = _("No fact with given PK found.")
            self.store.logger.error(message)
            raise KeyError(message)
        if not raw:
            # Explain: Why is as_hamster optionable, when act/cat/tag do it always?
            result = result.as_hamster(self.store)
        self.store.logger.debug(_("Returning {!r}.".format(result)))
        return result

    # ***

    def _get_all(
        self,
        endless=False,
        partial=False,
        include_usage=False,
        count_results=False,
        since=None,
        until=None,
        # FIXME/2018-06-09: (lb): Implement deleted/hidden.
        deleted=False,
        search_term='',
        activity=False,
        category=False,
        sort_col='',
        sort_order='',
        raw=False,
        exclude_ongoing=None,
        # (lb): IMPOSSIBLE_BRANCH: We should always be able to preload tags
        # (eager loading), which is a lot quicker than lazy-loading tags,
        # especially when exporting all Facts. I.e., when eager loading,
        # there's only one SELECT call; but if lazy loading, there'd be one
        # SELECT to get all the Facts, and then one SELECT to get the tags
        # for each Fact; inefficient!). In any case, if there are problems
        # with pre-loading, you can flip this switch to sample the other
        # behavior, which is SQLAlchemy's "default", which is to lazy-load.
        lazy_tags=False,
        # kwargs: limit, offset
        **kwargs
    ):
        """
        Return all facts within a given timeframe that match given search terms.

        ``get_all`` already took care of any normalization required.

        If no timeframe is given, return all facts.

        Args:
            deleted (boolean, optional): False to restrict to non-deleted
                Facts; True to find only those marked deleted; None to find
                all.
            since (datetime.datetime, optional):
                Match Facts more recent than a specific dates.
            until (datetime.datetime, optional):
                Match Facts older than a specific dates.
            search_term (str):
                Case-insensitive strings to match ``Activity.name`` or
                ``Category.name``.
            deleted (boolean, optional): False to restrict to non-deleted
                Facts; True to find only those marked deleted; None to find
                all.
            partial (bool):
                If ``False`` only facts which start *and* end within the
                timeframe will be considered. If ``True`` facts with
                either ``start``, ``end`` or both within the timeframe
                will be returned.
            order (string, optional): 'asc' or 'desc'; re: Fact.start.

        Returns:
            list: List of ``nark.Facts`` instances.

        Note:
            This method will *NOT* return facts that start before and end after
            (e.g. that span more than) the specified timeframe.
        """
        magic_tag_sep = '%%%%,%%%%'

        def _get_all_facts():
            message = _(
                'since: {} / until: {} / srch_term: {} / srt_col: {} / srt_ordr: {}'
                .format(since, until, search_term, sort_col, sort_order)
            )
            self.store.logger.debug(message)

            query = self.store.session.query(AlchemyFact)

            query, span_col = _get_all_prepare_span_col(query)

            query, tags_col = _get_all_prepare_tags_col(query)

            query = _get_all_prepare_joins(query)

            query = _get_all_filter_partial(query)

            query = _get_all_filter_by_activity(query)

            query = _get_all_filter_by_category(query)

            query = _get_all_filter_by_search_term(query)

            query = query_apply_true_or_not(query, AlchemyFact.deleted, deleted)

            query = _get_all_filter_by_ongoing(query)

            query = _get_all_order_by(query, span_col, tags_col)

            query = query_apply_limit_offset(query, **kwargs)

            query = _get_all_with_entities(query, span_col, tags_col)

            self.store.logger.debug(_('query: {}'.format(str(query))))

            if count_results:
                results = query.count()
            else:
                # Profiling: 2018-07-15: (lb): ~ 0.120 s. to fetch latest of 20K Facts.
                records = query.all()
                results = _process_results(records, span_col, tags_col)

            return results

        def _process_results(records, span_col, tags_col):
            if span_col is None and tags_col is None:
                if raw:
                    return records
                else:
                    return [fact.as_hamster(self.store) for fact in records]

            results = []
            for fact, *cols in records:
                new_tags = None
                if not lazy_tags:
                    tags = cols.pop()
                    new_tags = tags.split(magic_tag_sep) if tags else []

                if not raw:
                    new_fact = fact.as_hamster(self.store, new_tags)
                else:
                    if new_tags:
                        fact.tags = new_tags
                    new_fact = fact

                if len(cols):
                    results.append((new_fact, *cols))
                else:
                    results.append(new_fact)
            return results

        # ***

        def _get_all_prepare_tags_col(query):
            if lazy_tags:
                return query, None
            # (lb): Always include tags. We could let SQLAlchemy lazy load,
            # but this can be slow. E.g., on 15K Facts, calling fact.tags on
            # each -- triggering lazy load -- takes 7 seconds on my machine.
            # As opposed to 0 seconds (rounded down) when preloading tags.
            tags_col = func.group_concat(
                AlchemyTag.name, magic_tag_sep,
            ).label("facts_tags")
            query = query.add_columns(tags_col)
            query = query.outerjoin(fact_tags)
            query = query.outerjoin(AlchemyTag)
            query = query.group_by(AlchemyFact.pk)
            # (lb): 2019-01-22: Old comment re: joinedload. Leaving here as
            # documentation in case I try using joinedload again in future.
            #   # FIXME/2018-06-25: (lb): Not quite sure this'll work...
            #   # http://docs.sqlalchemy.org/en/latest/orm/loading_relationships.html
            #   #   # joined-eager-loading
            #   from sqlalchemy.orm import joinedload
            #   # 2019-01-22: Either did not need, or did not work, !remember which!
            #   query = query.options(joinedload(AlchemyFact.tags))
            return query, tags_col

        def _get_all_prepare_span_col(query):
            if not include_usage:
                return query, None

            span_col = (
                func.julianday(AlchemyFact.end) - func.julianday(AlchemyFact.start)
            ).label('span')
            query = self.store.session.query(AlchemyFact, span_col)
            return query, span_col

        def _get_all_prepare_joins(query):
            if (
                include_usage
                or (activity is not False)
                or (category is not False)
                or search_term
            ):
                query = query.outerjoin(AlchemyFact.activity)  # Same as: AlchemyActivity
            if (category is not False) or search_term:
                query = query.outerjoin(AlchemyCategory)
            return query

        def _get_all_filter_partial(query):
            fmt_since = self._get_sql_datetime(since) if since else None
            fmt_until = self._get_sql_datetime(until) if until else None
            if partial:
                query = _get_partial_overlaps(query, fmt_since, fmt_until)
            else:
                query = _get_complete_overlaps(query, fmt_since, fmt_until, endless)
            return query

        def _get_partial_overlaps(query, since, until):
            """Return all facts where either start or end falls within the timeframe."""
            if since and not until:
                # (lb): Checking AlchemyFact.end >= since is sorta redundant,
                # because AlchemyFact.start >= since should guarantee that.
                query = query.filter(
                    or_(
                        func.datetime(AlchemyFact.start) >= since,
                        func.datetime(AlchemyFact.end) >= since,
                    ),
                )
            elif not since and until:
                # (lb): Checking AlchemyFact.start <= until is sorta redundant,
                # because AlchemyFact.end <= until should guarantee that.
                query = query.filter(
                    or_(
                        func.datetime(AlchemyFact.start) <= until,
                        func.datetime(AlchemyFact.end) <= until,
                    ),
                )
            elif since and until:
                query = query.filter(or_(
                    and_(
                        func.datetime(AlchemyFact.start) >= since,
                        func.datetime(AlchemyFact.start) <= until,
                    ),
                    and_(
                        func.datetime(AlchemyFact.end) >= since,
                        func.datetime(AlchemyFact.end) <= until,
                    ),
                ))
            else:
                pass
            return query

        def _get_complete_overlaps(query, since, until, endless=False):
            """Return all facts with start and end within the timeframe."""
            if since:
                query = query.filter(func.datetime(AlchemyFact.start) >= since)
            if until:
                query = query.filter(func.datetime(AlchemyFact.end) <= until)
            elif endless:
                query = query.filter(AlchemyFact.end == None)  # noqa: E711
            return query

        def _get_all_filter_by_activity(query):
            if activity is False:
                return query

            if activity:
                if activity.pk:
                    query = query.filter(AlchemyActivity.pk == activity.pk)
                else:
                    query = query.filter(
                        func.lower(AlchemyActivity.name) == func.lower(activity.name)
                    )
            else:
                query = query.filter(AlchemyFact.activity == None)  # noqa: E711
            return query

        def _get_all_filter_by_category(query):
            if category is False:
                return query
            if category:
                if category.pk:
                    query = query.filter(AlchemyCategory.pk == category.pk)
                else:
                    query = query.filter(
                        func.lower(AlchemyCategory.name) == func.lower(category.name)
                    )
            else:
                query = query.filter(AlchemyFact.category == None)  # noqa: E711
            return query

        def _get_all_filter_by_search_term(query):
            if search_term:
                query = _filter_search_term(query)
            return query

        def _filter_search_term(query):
            """
            Limit query to facts that match the search terms.

            Terms are matched against ``Category.name`` and ``Activity.name``.
            The matching is not case-sensitive.
            """
            # FIXME/2018-06-09: (lb): Now with activity and category filters,
            #   search_term makes less sense. Unless we apply to all parts?
            #   E.g., match tags, and match description.
            query = query.filter(
                or_(
                    AlchemyActivity.name.ilike('%{}%'.format(search_term)),
                    AlchemyCategory.name.ilike('%{}%'.format(search_term)),
                )
            )

            return query

        def _get_all_filter_by_ongoing(query):
            if not exclude_ongoing:
                return query
            if exclude_ongoing:
                query = query.filter(AlchemyFact.end != None)  # noqa: E711
            return query

        # FIXME/2018-06-09: (lb): DRY: Combine each manager's _get_all_order_by.
        def _get_all_order_by(query, span_col=None, tags_col=None):
            direction = desc if sort_order == 'desc' else asc
            if sort_col == 'start':
                direction = desc if not sort_order else direction
                query = self._get_all_order_by_times(query, direction)
            elif sort_col == 'time':
                assert include_usage and span_col is not None
                direction = desc if not sort_order else direction
                query = query.order_by(direction(span_col))
            elif sort_col == 'activity':
                query = query.order_by(direction(AlchemyActivity.name))
                query = query.order_by(direction(AlchemyCategory.name))
            elif sort_col == 'category':
                query = query.order_by(direction(AlchemyCategory.name))
                query = query.order_by(direction(AlchemyActivity.name))
            else:
                # Meh. Rather than make a custom --order for each command,
                # just using the same big list. So 'activity', 'category',
                # etc., are acceptable here, if not simply ignored.
                assert sort_col in ('', 'name', 'tag', 'fact')
                direction = desc if not sort_order else direction
                query = self._get_all_order_by_times(query, direction)
            return query

        def _get_all_with_entities(query, span_col, tags_col):
            columns = []

            if span_col is not None:
                # Throw in the count column, which act/cat/tag fetch, so we can
                # use the same utility functions (that except a count column).
                static_count = '1'
                columns.append(static_count)
                columns.append(span_col)

            if tags_col is not None:
                assert not lazy_tags
                columns.append(tags_col)

            query = query.with_entities(AlchemyFact, *columns)

            return query

        # ***

        return _get_all_facts()

    # ***

    def _get_all_order_by_times(self, query, direction, fact=None):
        if fact and not fact.unstored:
            query = query.filter(AlchemyFact.pk != fact.pk)

        # Include end so that momentaneous Facts are sorted properly.
        # - And add PK, too, so momentaneous Facts are sorted predictably.
        query = query.order_by(
            direction(AlchemyFact.start),
            direction(AlchemyFact.end),
            direction(AlchemyFact.pk),
        )
        return query

    # ***

    def _get_sql_datetime(self, datetm):
        # Be explicit with the format used by the SQL engine, otherwise,
        #   e.g., and_(AlchemyFact.start > start) might match where
        #   AlchemyFact.start == start. In the case of SQLite, the stored
        #   date will be translated with the seconds, even if 0, e.g.,
        #   "2018-06-29 16:32:00", but the datetime we use for the compare
        #   gets translated without, e.g., "2018-06-29 16:32". And we
        #   all know that "2018-06-29 16:32:00" > "2018-06-29 16:32".
        # See also: func.datetime(AlchemyFact.start/end).
        cmp_fmt = '%Y-%m-%d %H:%M:%S'
        text = datetm.strftime(cmp_fmt)
        return text

    # ***

    def starting_at(self, fact):
        """
        Return the fact starting at the moment in time indicated by fact.start.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``start`` set.

        Returns:
            nark.Fact: The found Fact, or None if none found.

        Raises:
            ValueError: If more than one Fact found at given time.
        """
        query = self.store.session.query(AlchemyFact)

        if fact.start is None:
            raise ValueError('No `start` for starting_at(fact).')

        start_at = self._get_sql_datetime(fact.start)
        condition = and_(func.datetime(AlchemyFact.start) == start_at)

        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712

        query = query.filter(condition)

        query = self._get_all_order_by_times(query, asc, fact=fact)

        self.store.logger.debug(_('fact: {} / query: {}'.format(fact, str(query))))

        n_facts = query.count()
        if n_facts > 1:
            message = 'More than one fact found starting at "{}": {} facts found'.format(
                fact.start, n_facts
            )
            raise ValueError(message)

        found = query.one_or_none()
        found_fact = found.as_hamster(self.store) if found else None
        return found_fact

    # ***

    def ending_at(self, fact):
        """
        Return the fact ending at the moment in time indicated by fact.end.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``end`` set.

        Returns:
            nark.Fact: The found Fact, or None if none found.

        Raises:
            ValueError: If more than one Fact found at given time.
        """
        query = self.store.session.query(AlchemyFact)

        if fact.end is None:
            raise ValueError('No `end` for ending_at(fact).')

        end_at = self._get_sql_datetime(fact.end)
        condition = and_(func.datetime(AlchemyFact.end) == end_at)

        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712

        query = query.filter(condition)

        query = self._get_all_order_by_times(query, desc, fact=fact)

        self.store.logger.debug(_('fact: {} / query: {}'.format(fact, str(query))))

        n_facts = query.count()
        if n_facts > 1:
            message = 'More than one fact found ending at "{}": {} facts found'.format(
                fact.end, n_facts,
            )
            raise ValueError(message)

        found = query.one_or_none()
        found_fact = found.as_hamster(self.store) if found else None
        return found_fact

    # ***

    def antecedent(self, fact=None, ref_time=None):
        """
        Return the Fact immediately preceding the indicated Fact.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``start`` set.

            ref_time (datetime.datetime):
                In lieu of fact, pass the datetime to reference.

        Returns:
            nark.Fact: The antecedent Fact, or None if none found.

        Raises:
            ValueError: If neither ``start`` nor ``end`` is set on fact.
        """
        query = self.store.session.query(AlchemyFact)

        if fact is not None:
            if fact.end and isinstance(fact.end, datetime):
                ref_time = fact.end
            elif fact.start and isinstance(fact.start, datetime):
                ref_time = fact.start
        if not isinstance(ref_time, datetime):
            raise ValueError(_('No reference time for antecedent(fact).'))

        ref_time = self._get_sql_datetime(ref_time)

        before_ongoing_fact_start = and_(
            AlchemyFact.end == None,  # noqa: E711
            # Except rather than <=, use less than, otherwise
            # penultimate_fact.antecedent might find the ultimate
            # fact, if that final fact is ongoing.
            #   E.g., considering
            #     fact  1: time-a to time-b
            #     ...
            #     fact -2: time-x to time-y
            #     fact -1: time-y to <now>
            #   antecedent of fact -2 should check time-y < time-y and
            #   not <= otherwise antecedent of fact -2 would be fact -1.
            #   (The subsequent function will see it, though, as it
            #   looks for AlchemyFact.start >= ref_time.)
            func.datetime(AlchemyFact.start) < ref_time,
        )

        if fact is None or fact.pk is None:
            before_closed_fact_end = and_(
                AlchemyFact.end != None,  # noqa: E711
                # Special case for ongoing fact (check its start).
                # Be most inclusive and compare against facts' ends.
                func.datetime(AlchemyFact.end) <= ref_time,
            )
        else:
            before_closed_fact_end = and_(
                AlchemyFact.end != None,  # noqa: E711
                or_(
                    func.datetime(AlchemyFact.end) < ref_time,
                    and_(
                        func.datetime(AlchemyFact.end) == ref_time,
                        AlchemyFact.pk != fact.pk,
                    ),
                ),
            )

        condition = or_(
            before_ongoing_fact_start,
            before_closed_fact_end,
        )

        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712

        query = query.filter(condition)

        # Exclude fact.pk from results.
        query = self._get_all_order_by_times(query, desc, fact=fact)

        query = query.limit(1)

        self.store.logger.debug(_(
            'fact: {} / ref_time: {} / query: {}'
            .format(fact, ref_time, str(query))
        ))

        found = query.one_or_none()
        found_fact = found.as_hamster(self.store) if found else None
        return found_fact

    # ***

    def subsequent(self, fact=None, ref_time=None):
        """
        Return the Fact immediately following the indicated Fact.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``end`` set.

            ref_time (datetime.datetime):
                In lieu of fact, pass the datetime to reference.

        Returns:
            nark.Fact: The subsequent Fact, or None if none found.

        Raises:
            ValueError: If neither ``start`` nor ``end`` is set on fact.
        """
        query = self.store.session.query(AlchemyFact)

        if fact is not None:
            if fact.start and isinstance(fact.start, datetime):
                ref_time = fact.start
            elif fact.end and isinstance(fact.end, datetime):
                ref_time = fact.end
        if ref_time is None:
            raise ValueError(_('No reference time for subsequent(fact).'))

        ref_time = self._get_sql_datetime(ref_time)

        if fact is None or fact.pk is None:
            condition = and_(func.datetime(AlchemyFact.start) >= ref_time)
        else:
            condition = or_(
                func.datetime(AlchemyFact.start) > ref_time,
                and_(
                    func.datetime(AlchemyFact.start) == ref_time,
                    AlchemyFact.pk != fact.pk,
                ),
            )

        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712

        query = query.filter(condition)

        # Exclude fact.pk from results.
        query = self._get_all_order_by_times(query, asc, fact=fact)

        query = query.limit(1)

        self.store.logger.debug(_(
            'fact: {} / ref_time: {} / query: {}'
            .format(fact, ref_time, str(query))
        ))

        found = query.one_or_none()
        found_fact = found.as_hamster(self.store) if found else None
        return found_fact

    # ***

    def strictly_during(self, since, until, result_limit=1000):
        """
        Return the fact(s) strictly contained within a since and until time.

        Args:
            since (datetime.datetime):
                Start datetime of facts to find.

            until (datetime.datetime):
                End datetime of facts to find.

            result_limit (int):
                Maximum number of facts to find, else log warning message.

        Returns:
            list: List of ``nark.Facts`` instances.
        """
        query = self.store.session.query(AlchemyFact)

        condition = and_(
            func.datetime(AlchemyFact.start) >= self._get_sql_datetime(since),
            or_(
                and_(
                    AlchemyFact.end != None,  # noqa: E711
                    func.datetime(AlchemyFact.end) <= self._get_sql_datetime(until),
                ),
                and_(
                    AlchemyFact.end == None,  # noqa: E711
                    func.datetime(AlchemyFact.start) <= self._get_sql_datetime(until),
                ),
            ),
        )

        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712

        query = query.filter(condition)

        query = self._get_all_order_by_times(query, asc)

        self.store.logger.debug(_(
            'since: {} / until: {} / query: {}'
            .format(since, until, str(query))
        ))

        # LATER: (lb): We'll let the client ask for as many records as they
        # want. But we might want to offer ways to deal more gracefully with
        # it, like via pagination; or a fetch_one callback, so that only item
        # gets loaded in memory at a time, rather than everything. For now, we
        # can at least warn, I suppose.
        during_count = query.count()
        if during_count > result_limit:
            # (lb): hamster-lib would `raise OverflowError`,
            # but that seems drastic.
            message = (_(
                'This is your alert that lots of Facts were found between '
                'the two dates specified: found {}.'
                .factor(during_count)
            ))
            self.store.logger.warning(message)

        facts = query.all()
        found_facts = [fact.as_hamster(self.store) for fact in facts]
        return found_facts

    # ***

    def surrounding(self, fact_time, inclusive=False):
        """
        Return the fact(s) at the given moment in time.
        Note that this excludes a fact that starts or ends at this time.
        (See antecedent and subsequent for finding those facts.)

        Args:
            fact_time (datetime.datetime):
                Time of fact(s) to match.

        Returns:
            list: List of ``nark.Facts`` instances.

        Raises:
            ValueError: If more than one Fact found at given time.
        """
        query = self.store.session.query(AlchemyFact)

        cmp_time = self._get_sql_datetime(fact_time)

        if not inclusive:
            condition = and_(
                func.datetime(AlchemyFact.start) < cmp_time,
                # Find surrounding complete facts, or the ongoing fact.
                or_(
                    AlchemyFact.end == None,  # noqa: E711
                    func.datetime(AlchemyFact.end) > cmp_time,
                ),
            )
        else:
            condition = and_(
                func.datetime(AlchemyFact.start) <= cmp_time,
                # Find surrounding complete facts, or the ongoing fact.
                or_(
                    AlchemyFact.end == None,  # noqa: E711
                    func.datetime(AlchemyFact.end) >= cmp_time,
                ),
            )

        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712

        query = query.filter(condition)

        query = self._get_all_order_by_times(query, asc)

        self.store.logger.debug(_(
            'fact_time: {} / query: {}'.format(
                fact_time, str(query)
            )
        ))

        if not inclusive:
            n_facts = query.count()
            if n_facts > 1:
                message = 'Broken time frame found at "{}": {} facts found'.format(
                    fact_time, n_facts
                )
                raise ValueError(message)

        facts = query.all()
        found_facts = [fact.as_hamster(self.store) for fact in facts]
        return found_facts

    # ***

    def endless(self):
        """
        Return any facts without a fact.start or fact.end.

        Args:
            <none>

        Returns:
            list: List of ``nark.Facts`` instances.
        """
        query = self.store.session.query(AlchemyFact)

        # NOTE: (lb): Use ==/!=, not `is`/`not`, b/c SQLAlchemy
        #       overrides ==/!=, not `is`/`not`.
        condition = or_(AlchemyFact.start == None, AlchemyFact.end == None)  # noqa: E711
        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712

        query = query.filter(condition)

        self.store.logger.debug(_('query: {}'.format(str(query))))

        facts = query.all()
        found_facts = [fact.as_hamster(self.store) for fact in facts]
        return found_facts

